{
  +---------------------------------------------------------------------------+
  |                       Unit SPRITEZ.PAS
  |
  +---------------------------------------------------------------------------+
  Author      :  Jean-Yves QUEINEC  j.y.q@wanadoo.fr
  Object      :  sprites unit for Qpuzzle
  +---------------------------------------------------------------------------+
}
UNIT Spritez;
{ ** optimizations ** }
{$R-} { range checking off }
{$Q-} { stack checking off }
{$I-} { i/o checking off }

INTERFACE

USES
{$IF CompilerVersion > 28} // Delphi XE7
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Classes, System.Types, System.UITypes, Vcl.Graphics;
{$ELSE}
  Windows, Messages, SysUtils, Classes, Graphics;
{$IFEND}

Const
  AmLeft = 1;
  AmTop = 1;
  AmRight = 1;
  AmBottom = 1;

TYPE
  TToon = CLASS(TObject)
  Private
    Dmaskcolor: Tcolor;
    Drect: Trect;
    DWidth: Integer;
    DHeight: Integer;
    Nbimages: Integer;
    Images: Tlist;
    Masks: Tlist;
    Origpt: Tpoint;

    Constructor Create;
    Procedure Free;
    Procedure LoadToon(Wbmp: Tbitmap; Nb, Row, Col, W, H, Separator: Integer; Maskcolor: Tcolor);
  End;

  TSprite = CLASS(TObject)
  Private
    Numero: Integer;
    Next: TSprite;
    Prior: TSprite;
    Priority: Integer;
    LoopCtr: Integer;
    Loopmax: Integer;
    DelayCtr: Integer;
    Sprtag: Integer;
    SprState: Byte;
    Filler: Byte;

    Toon: TToon;
    CurImage: Integer;
    Animloop: Integer;
    Animctr: Integer;
    Animsens: Integer;
    Animstyle: Integer;

    X: Integer;
    Y: Integer;
    Dcx: Integer;
    Dcy: Integer;
    Xx: Single;
    Yy: Single;
    Vx: Single;
    Vy: Single;
    Oldvx: Single;
    Oldvy: Single;
    Ax: Single;
    Ay: Single;
    Kvx: Single;
    Kvy: Single;

    Rcbounds: Trect;
    Rclimite: Trect;
    RcImg: Trect;
    RcMano: Trect;
    RcModif: Trect;

    Rebond: Trect;
    ReBrect: Trect;

    // AmGauche: Single;   // translate and moved
    // AmHaut: Single;
    // AmDroite: Single;
    // AmBas: Single;

    CONSTRUCTOR Create;
    Procedure Calculpos; VIRTUAL;
    Procedure Adjustlimits;
  End;

  { ----------------------------------------------------------------------- }
  { CLASS TPOOL   manages sets of sprites }
  { ----------------------------------------------------------------------- }

TYPE
  TPool = CLASS(TObject)
  Private
    SpriteFirst: TSprite;
    SpriteLast: TSprite;
    Tableau: Tlist;
    Palbum: Pointer;
    Style: Integer;
    Albumdraw: Boolean;

    CONSTRUCTOR Create(MaxS: Integer);
    Procedure Free;
    Procedure ChainSprite(S: TSprite);
    Procedure Chainlast(S: TSprite);
    Procedure DechainSprite(S: TSprite);
    Procedure PoolRemove; // efface
    Procedure PoolMove; // deplace
    Procedure PrepareCache;
    Procedure SpritesShow;
    Function SpriteCreate(N: Integer; SpriteType: Integer): Boolean;
    Function GetSprite(N: Integer): TSprite;
  Public
    Maxspr: Integer;
    Collisspr: Integer;
    Collisrect: Trect;
    PoolState: Integer;

    Function CreateSprite(N: Integer; Toonname: String): Boolean;
    Function SpriteExists(N: Integer): Boolean;
    Function SpriteCount: Integer;
    Procedure SpriteDestroy(N: Integer);
    Procedure SpriteErase(N: Integer);

    Procedure Animate;
    Procedure AnimateMono(N: Integer);

    Procedure SetLoopCtr(N, M: Integer);
    Function GetLoopCtr(N: Integer): Integer;
    Procedure SetDelayCtr(N, M: Integer);
    Function GetDelayCtr(N: Integer): Integer;
    Procedure Setpriority(N, P: Integer);
    Procedure Setprioritylast(N, P: Integer);
    Function Getpriority(N: Integer): Integer;
    Procedure SetActive(N: Integer; B: Boolean);
    Procedure SetVisible(N: Integer; B: Boolean);
    Function IsActive(N: Integer): Boolean;
    Function IsVisible(N: Integer): Boolean;
    Procedure Settag(N, Atag: Integer);
    Function Gettag(N: Integer): Integer;

    Procedure SetToon(N: Integer; Wname: STRING);
    Function GetToonName(N: Integer): String;
    Function GetToonW(N: Integer): Integer;
    Function GetToonH(N: Integer): Integer;
    Procedure SetAnim(N, Wcount: Integer);
    Procedure SetAnimStyle(N, Wstyle: Integer);
    Function GetCurimage(N: Integer): Integer;
    Procedure SetTrace(N: Integer; B: Boolean);

    Procedure SetLimits(N: Integer; Wrect: Trect);
    Procedure SetPos(N, Wx, Wy: Integer);
    Function Getposx(N: Integer): Integer;
    Function Getposy(N: Integer): Integer;
    Function Getdcx(N: Integer): Integer;
    Function Getdcy(N: Integer): Integer;
    Procedure SetV(N: Integer; Wvx, Wvy: Single);
    Procedure SetKV(N: Integer; Kvx, Kvy: Single);
    Procedure Stop(N: Integer);
    Function Getvx(N: Integer): Single;
    Function Getvy(N: Integer): Single;
    Function SameXdirection(N: Integer): Boolean;
    Function SameYdirection(N: Integer): Boolean;
    Procedure Setaccel(N: Integer; Wax, Way: Single);
    Function Getax(N: Integer): Single;
    Function Getay(N: Integer): Single;
    Function Getaxvnull(N, X1: Integer): Single;
    Function Getayvnull(N, Y1: Integer): Single;
    Procedure Gotoxy(N, Wx, Wy, Loops: Integer);
    Function PtinSprite(Pt: Tpoint; Marge: Integer): Integer;
    Function Getoffsetx(N: Integer): Integer;
    Function Getoffsety(N: Integer): Integer;
    Procedure Setoffset(N, Wx, Wy: Integer);
  End;

  { ----------------------------------------------------------------------- }
  { CLASS TALBUM manages Toons Decors and pools }
  { ----------------------------------------------------------------------- }

TYPE
  TAlbum = CLASS(TObject)
  Private
    Cancan: Tcanvas;
    BmpFond: Tbitmap;
    BmpMano: Tbitmap;
    Fondwidth: Integer;
    Fondheight: Integer;
    Toons: Tstringlist;
    Pools: Tlist;
    Flagdecor: Boolean;
    Function GetToon(Wname: STRING): TToon;
    Function GetAlbumToonName(AToon: TToon): String;
    Procedure Bmpfree;
    Procedure ToonsFree;
    Procedure PoolsFree;
  Public
    CONSTRUCTOR Create(Wcanvas: Tcanvas; Bitmap: Tbitmap);
    Procedure Free;
    Procedure FreePoolsToons;
    Function CreatePool(Maxsprites: Integer): TPool;
    Procedure FreePool(Wpool: TPool);

    Function GetAreaBmp: Tbitmap;
    Procedure SetAreaBmp(Wbmp: Tbitmap);
    Function GetArearect: Trect;
    Function GetAreawidth: Integer;
    Function GetAreaHeight: Integer;
    Procedure DrawArea(Wrect: Trect);

    Procedure FreeToon(Wname: STRING);
    Procedure CreateToon(Wname: STRING; Wbmp: Tbitmap; Nb, Row, Col, W, H, Separator: Integer; Maskcolor: Tcolor);
    Function GetMaskBitmap(Wname: String; N: Integer): Tbitmap;
    Function GetImgBitmap(Wname: String; N: Integer): Tbitmap;
    Function Gettoonwidth(Wname: String): Integer;
    Function GettoonHeight(Wname: String): Integer;
    Procedure SettoonCenter(Wname: String; Acx, Acy: Integer);
  End;

Function Rndi(A, B: Integer): Integer;
Function Rnd(A, B: Single): Single;
Function Zsin(A: Integer): Single;
Function Zcos(A: Integer): Single;
Function Distance(X1, Y1, X2, Y2: Integer): Integer;

IMPLEMENTATION

Uses Umain;

Const
  Maxrand = 1023;
  Actif = 0;
  Visible = 1;
  Trace = 2;
  Modif = 3;
  // Userdraw = 4;
  // Autoleader = 5;

Var
  Trandom: Array [0 .. Maxrand] OF Single;
  Irandom: Integer;
  Sinb: Array [0 .. 360] OF Single;
  Cosb: Array [0 .. 360] OF Single;

Procedure Log(Logregel: String);
Begin
  If Formmain <> Nil Then
    Formmain.Log(Logregel);
End;

Procedure Initrandom;
Const
  PiRad = Pi / 180;
Var
  I: Integer;
  A: Single;
Begin
  Log('& initrandom');
  Randomize;
  For I := 0 To Maxrand Do
    Trandom[I] := Random;
  For I := 0 To 360 Do
    Begin
      A := I * PiRad;
      Sinb[I] := Sin(A);
      Cosb[I] := Cos(A);
    End;
  Irandom := 0;
End;

Procedure SetBit(Position: Integer; Value: Byte; Var ChangeByte: Byte);
Var
  Bt: Byte;
Begin
  Bt := $01;
  Bt := Bt Shl Position;
  If Value = 1 Then
    ChangeByte := ChangeByte Or Bt
  Else
    Begin
      Bt := Bt Xor $FF;
      ChangeByte := ChangeByte And Bt;
    End;
End;

Function Getbit(Position: Integer; TestByte: Byte): Boolean;
Var
  Bt: Byte;
Begin
  Bt := $01;
  Bt := Bt Shl Position;
  Getbit := (Bt And TestByte) > 0;
End;

//Function CheckBetweenLimits(A: Integer): Integer;
//Begin
//  Result := A;
//  While Result > 360 Do
//    Dec(Result, 360);
//  While Result < 0 Do
//    Inc(Result, 360);
//End;

Function Zsin(A: Integer): Single;
Begin
  // Result := Sinb[CheckBetweenLimits(A)];
  While A > 360 Do
    Dec(A, 360);
  While A < 0 Do
    Inc(A, 360);
  Result := Sinb[A];
End;

Function Zcos(A: Integer): Single;
Begin
  // Result := Cosb[CheckBetweenLimits(A)];
  While A > 360 Do
    Dec(A, 360);
  While A < 0 Do
    Inc(A, 360);
  Result := Cosb[A];
End;

Function Rnd(A, B: Single): Single;
  Procedure Swap(Var A, B: Single);
  Var
    Temp: Single;
  Begin
    Temp := A;
    A := B;
    B := Temp;
  End;

Begin
  If B < A Then
    Swap(A, B);
  Result := A + (B - A) * Trandom[Irandom];
  Inc(Irandom);
  IF Irandom > Maxrand Then
    Irandom := 0;
End;

Function Rndi(A, B: Integer): Integer;
  Procedure Swap(Var A, B: Integer);
  Var
    Temp: Integer;
  Begin
    Temp := A;
    A := B;
    B := Temp;
  End;

Begin
  If B < A Then
    Swap(A, B);
  Result := A + Trunc((B - A + 1) * Trandom[Irandom]);
  Inc(Irandom);
  IF Irandom > Maxrand Then
    Irandom := 0;
END;

Function Distance(X1, Y1, X2, Y2: Integer): Integer;
Var
  I: Integer;
  A, R0, R1: Integer;
Begin
  Log('& distance');
  R0 := X2 - X1;
  R1 := Y2 - Y1;
  A := R0 * R0 + R1 * R1;
  R1 := (1 + A) DIV 2;
  FOR I := 1 TO 12 DO
    Begin
      R0 := R1;
      IF R0 > 0 Then
        R1 := (R0 + (A Div R0)) Div 2;
    End;
  Result := R1;
End;

{ ======================================================================
  Classe  TToon
  ====================================================================== }
CONSTRUCTOR TToon.Create;
Begin
  INHERITED Create;
  Dmaskcolor := Clblack;
  Drect := Rect(0, 0, 0, 0);
  DWidth := 0;
  DHeight := 0;
  Nbimages := 0;
  Images := Nil;
  Masks := Nil;
End;

Procedure TToon.Free;
Var
  I: Integer;
Begin
  FOR I := 0 TO Images.Count - 1 DO
    IF Assigned(Images.Items[I]) Then
      Tbitmap(Images.Items[I]).Free;
  FOR I := 0 TO Masks.Count - 1 DO
    IF Assigned(Masks.Items[I]) Then
      Tbitmap(Masks.Items[I]).Free;
  INHERITED Free;
End;

Procedure Domask(Bmpimage, Bmpmask: Tbitmap; Zw, Zh: Integer; Mcolo: Tcolor);
Var
  Svcolo: Tcolor;
Begin
  Log('Do Mask');
  Bmpmask.Monochrome := True;
  Svcolo := Bmpimage.Canvas.Brush.Color;
  Bmpimage.Canvas.Brush.Color := Mcolo;
  Bmpmask.Canvas.Draw(0, 0, Bmpimage);
  IF Mcolo <> Clblack Then
    Begin
      With Bmpimage.Canvas Do
        Begin
          Brush.Color := Mcolo;
          CopyMode := Cmsrcinvert;
          CopyRect(Rect(0, 0, Zw, Zh), Bmpmask.Canvas, Rect(0, 0, Zw, Zh));
          CopyMode := Cmsrccopy;
        End;
    End;
  Bmpimage.Canvas.Brush.Color := Svcolo;
End;

//  separator = 1   +------+------+     separator = 0  XXXXXXWWWWWW
//                  |XXXXXX|WWWWWW|                    XXXXXXWWWWWW
//                  |XXXXXX|WWWWWW|                    WWWWWW
//                  +------+------+                    WWWWWW
//                  |WWWWWW|      |
//                  |WWWWWW|      |
//                  +------+------+

Procedure TToon.LoadToon(Wbmp: Tbitmap; Nb, Row, Col, W, H, Separator: Integer; Maskcolor: Tcolor);
Var
  Bmpi: Tbitmap;
  Bmpm: Tbitmap;
  Ctr: Integer;
  Ix: Integer;
  Iy: Integer;
  Colonne: Integer;
  OrigRect: Trect;
Begin
  Log('& loadtoon');
  Dmaskcolor := Maskcolor;
  DWidth := W;
  DHeight := H;
  IF Nb > Row * Col Then
    Nb := Row * Col;
  Nbimages := Nb;
  Images := Tlist.Create;
  Images.Capacity := Nb;
  Masks := Tlist.Create;
  Masks.Capacity := Nb;
  IF Separator < 0 Then
    Separator := 0;
  Ix := Separator;
  Iy := Separator;
  Colonne := 1;
  FOR Ctr := 1 TO Nbimages DO
    Begin
      Bmpi := Tbitmap.Create;
      Bmpm := Tbitmap.Create;
      Try
        Bmpi.Width := DWidth;
        Bmpi.Height := DHeight;
        Bmpi.Palette := Wbmp.Palette;
        Bmpi.Canvas.Brush.Color := Dmaskcolor;
        Bmpi.Canvas.Fillrect(Rect(0, 0, DWidth, DHeight));
        OrigRect := Bounds(Ix, Iy, W, H);
        Bmpi.Canvas.CopyRect(Rect(0, 0, W, H), Wbmp.Canvas, OrigRect);

        Bmpm.Width := DWidth;
        Bmpm.Height := DHeight;
        Domask(Bmpi, Bmpm, DWidth, DHeight, Dmaskcolor);
        Ix := Ix + W + Separator;
        Inc(Colonne);
        IF Colonne > Col Then
          Begin
            Iy := Iy + H + Separator;
            Colonne := 1;
            Ix := Separator;
          End;
      Finally
        IF Assigned(Bmpi) Then
          Images.Add(Bmpi);
        // Bmpi.Free;
        IF Assigned(Bmpm) Then
          Masks.Add(Bmpm);
        // Bmpm.Free;
      End;
    End;
End;

{ ==========================================================================
  Class  TAlbum
  ========================================================================== }
CONSTRUCTOR TAlbum.Create(Wcanvas: Tcanvas; Bitmap: Tbitmap);
Begin
  Log('& album create');
  INHERITED Create;
  Cancan := Wcanvas;
  Toons := Tstringlist.Create;
  Toons.Capacity := 1024;
  Pools := Tlist.Create;
  BmpFond := Nil;
  BmpMano := Nil;
  SetAreaBmp(Bitmap);
  Flagdecor := False;
End;

Procedure TAlbum.ToonsFree;
Var
  I: Integer;
Begin
  IF Assigned(Toons) Then
    Begin
      FOR I := 0 TO Toons.Count - 1 DO
        TToon(Toons.Objects[I]).Free;
      // Toons.Objects[I] := Nil;
      Toons.Free;
    End;
End;

Procedure TAlbum.PoolsFree;
Var
  I: Integer;
Begin
  IF Assigned(Pools) Then
    Begin
      FOR I := 0 TO Pools.Count - 1 DO
        TPool(Pools.Items[I]).Free;
      Pools.Free;
    End;
End;

Procedure TAlbum.Free;
Begin
  Log('Album free');
  PoolsFree;
  ToonsFree;
  Bmpfree;
  INHERITED Free;
  // Self := Nil;
End;

Procedure TAlbum.FreePoolsToons;
Begin
  Log('Poolstones free');
  PoolsFree;
  ToonsFree;
End;

Function TAlbum.CreatePool(Maxsprites: Integer): TPool;
Var
  Pp: TPool;
Begin
  Log('Create pool');
  Pp := TPool.Create(Maxsprites);
  IF Assigned(Pp) Then
    Begin
      Pp.Palbum := Self;
      Pools.Add(Pp);
      Result := Pp;
      Pp.Style := 0;
      Pp.Albumdraw := False;
    End
  Else
    Result := Nil;
End;

Procedure TAlbum.FreePool(Wpool: TPool);
Begin
  Log('Free pool');
  IF Pools.Indexof(Wpool) <> -1 Then
    Begin
      Wpool.Free;
      Pools.Remove(Wpool);
    End;
  // Wpool := Nil;
End;

Procedure TAlbum.SetAreaBmp(Wbmp: Tbitmap);
Begin
  Log('Set areabmp');
  Bmpfree;

  BmpFond := Tbitmap.Create;
  BmpFond.Width := Wbmp.Width;
  BmpFond.Height := Wbmp.Height;
  BmpFond.Canvas.Draw(0, 0, Wbmp);
  BmpFond.Palette := Wbmp.Palette;
  Wbmp.Releasepalette;
  // bmpfond.SaveToFile('bmpfond.bmp');

  BmpMano := Tbitmap.Create;
  // BmpMano := BmpFond;
  BmpMano.Width := Wbmp.Width;
  BmpMano.Height := Wbmp.Height;
  BmpMano.Canvas.Draw(0, 0, Wbmp);
  BmpMano.Palette := Wbmp.Palette;
  // bmpmano.SaveToFile('bmpmano.bmp');
  Wbmp.Releasepalette;

  Fondwidth := Wbmp.Width;
  Fondheight := Wbmp.Height;
End;

Function TAlbum.GetAreaBmp: Tbitmap;
Begin
  Result := BmpFond;
End;

Function TAlbum.GetArearect: Trect;
Begin
  Result := Rect(0, 0, Fondwidth, Fondheight);
  // Result.Left := 0;
  // Result.Top := 0;
  // Result.Right := Fondwidth;
  // Result.Bottom := Fondheight;
End;

Function TAlbum.GetAreawidth: Integer;
Begin
  Result := Fondwidth;
End;

Function TAlbum.GetAreaHeight: Integer;
Begin
  Result := Fondheight;
End;

Procedure TAlbum.DrawArea(Wrect: Trect);
Var
  Oldpal: Hpalette;
Begin
  Log('Draw area');
  BmpMano.Canvas.CopyMode := Cmsrccopy;
  BmpMano.Canvas.Draw(0, 0, BmpFond);
  Oldpal := Selectpalette(Cancan.Handle, BmpFond.Palette, False);
  Realizepalette(Cancan.Handle);
  Cancan.Stretchdraw(Wrect, BmpMano);
  Selectpalette(Cancan.Handle, Oldpal, False);
End;

Procedure TAlbum.CreateToon(Wname: STRING; Wbmp: Tbitmap; Nb, Row, Col, W, H, Separator: Integer; Maskcolor: Tcolor);
Var
  D: TToon;
Begin
  Log('Create toon (' + Wname + ' =nb:' + Inttostr(Nb) + ', ' + Inttostr(Row) + ',' + Inttostr(Col) + ' ');
  D := TToon.Create;
  D.LoadToon(Wbmp, Nb, Row, Col, W, H, Separator, Maskcolor);
  Toons.Addobject(Uppercase(Wname), D);
End;

Function TAlbum.GetMaskBitmap(Wname: String; N: Integer): Tbitmap;
Var
  I: Integer;
  D: TToon;
Begin
  Result := Nil;
  I := Toons.Indexof(Uppercase(Wname));
  IF I >= 0 Then
    Begin
      D := TToon(Toons.Objects[I]);
      IF N > D.Masks.Count - 1 Then
        Exit;
      Result := Tbitmap(D.Masks[N]);
    End;
End;

Function TAlbum.GetImgBitmap(Wname: String; N: Integer): Tbitmap;
Var
  I: Integer;
  D: TToon;
Begin
  Result := Nil;
  I := Toons.Indexof(Uppercase(Wname));
  IF I >= 0 Then
    Begin
      D := TToon(Toons.Objects[I]);
      IF N > D.Images.Count - 1 Then
        Exit;
      Result := Tbitmap(D.Images[N]);
    End;
End;

Function TAlbum.Gettoonwidth(Wname: String): Integer;
Var
  I: Integer;
  D: TToon;
Begin
  Result := 8;
  I := Toons.Indexof(Uppercase(Wname));
  IF I >= 0 Then
    Begin
      D := TToon(Toons.Objects[I]);
      Result := D.DWidth;
    End;
End;

Function TAlbum.GettoonHeight(Wname: String): Integer;
Var
  I: Integer;
  D: TToon;
Begin
  Result := 8;
  I := Toons.Indexof(Uppercase(Wname));
  IF I >= 0 Then
    Begin
      D := TToon(Toons.Objects[I]);
      Result := D.DHeight;
    End;
End;

Procedure TAlbum.FreeToon(Wname: STRING);
Var
  I: Integer;
Begin
  I := Toons.Indexof(Uppercase(Wname));
  IF I >= 0 Then
    Begin
      TToon(Toons.Objects[I]).Free;
      Toons.Delete(I);
    End;
End;

Function TAlbum.GetToon(Wname: STRING): TToon;
Var
  I: Integer;
Begin
  I := Toons.Indexof(Uppercase(Wname));
  IF I < 0 Then
    Result := Nil
  Else
    Result := TToon(Toons.Objects[I]);
End;

Function TAlbum.GetAlbumToonName(AToon: TToon): String;
Var
  I: Integer;
Begin
  Log('Get album toonname');
  Result := '';
  For I := 0 To Toons.Count - 1 Do
    Begin
      IF Toons.Objects[I] = AToon Then
        Begin
          Result := Toons[I];
          Break;
        End;
    End;
End;

Procedure TAlbum.SettoonCenter(Wname: String; Acx, Acy: Integer);
Var
  D: TToon;
Begin
  D := GetToon(Wname);
  If D = Nil Then
    Exit;
  D.Origpt.X := Acx;
  D.Origpt.Y := Acy;
End;

Procedure TAlbum.Bmpfree;
Begin
  If Assigned(BmpFond) Then
    BmpFond.Free;
  If Assigned(BmpMano) Then
    BmpMano.Free;
End;

{ ==========================================================================
  Classe  TSprite
  ========================================================================== }
CONSTRUCTOR TSprite.Create;
Begin
  INHERITED Create;
  Next := Nil;
  Priority := 0;
  SprState := $00;
  SetBit(Actif, 1, SprState);
  SetBit(Modif, 1, SprState);
  Toon := Nil;
  LoopCtr := 0;
  Loopmax := 0;
  DelayCtr := 0;
  CurImage := 0;
  Animloop := 10;
  Animctr := 1;
  Animsens := 1;
  Animstyle := 0;
  Vx := 0;
  Vy := 0;
  Ax := 0;
  Ay := 0;
  Kvx := 1;
  Kvy := 1;
  Rcbounds := Rect(0, 0, 0, 0);
  Rclimite := Rect(0, 0, 0, 0);
  RcImg := Rect(0, 0, 0, 0);
  RcMano := Rect(0, 0, 0, 0);
  RcModif := Rect(0, 0, 0, 0);
  Rebond := Rect(0, 0, 0, 0);
  ReBrect := Rect(0, 0, 0, 0);
  // AmLeft := 1;      moved
  // AmTop := 1;
  // AmRight := 1;
  // AmBottom := 1;
  Sprtag := 0;
  Filler := 0;
End;

Procedure TSprite.Adjustlimits;
Begin
  If Toon = Nil Then
    Exit;
  Rclimite := Rcbounds;
  IF Rebond.Left <> 1 Then
    Rclimite.Left := Rcbounds.Left - Toon.DWidth;
  IF Rebond.Top <> 1 Then
    Rclimite.Top := Rcbounds.Top - Toon.DHeight;
  IF Rebond.Right = 1 Then
    Rclimite.Right := Rcbounds.Right - Toon.DWidth;
  IF Rebond.Bottom = 1 Then
    Rclimite.Bottom := Rcbounds.Bottom - Toon.DHeight;
End;

Procedure TSprite.Calculpos;
  Procedure SetZero;
  Begin
    Vx := 0;
    Vy := 0;
    Ax := 0;
    Ay := 0;
    Kvx := 1;
    Kvy := 1
  End;

Begin
  Oldvx := Vx;
  Oldvy := Vy;
  Vx := Vx * Kvx + Ax;
  Vy := Vy * Kvy + Ay;
  Xx := Xx + Vx;
  Yy := Yy + Vy;
  X := Trunc(Xx);
  Y := Trunc(Yy);
  ReBrect := Rect(0, 0, 0, 0);
  IF X < Rclimite.Left Then
    Begin
      ReBrect.Left := 1;
      Case Rebond.Left OF
        0:
          Begin
            Xx := Rclimite.Right;
            X := Trunc(Xx);
          End;
        1:
          Begin
            Xx := Rclimite.Left;
            X := Trunc(Xx);
            IF Abs(Vx) < 0.0001 Then
              Vx := 0
            Else
              Vx := -Vx * AmLeft;
          End;
        2:
          Begin
            SetZero;
          End;
      End;
    END
  Else
    IF Y < Rclimite.Top Then
      Begin
        ReBrect.Top := 1;
        Case Rebond.Top OF
          0:
            Begin
              Yy := Rclimite.Bottom;
              Y := Trunc(Y);
            End;
          1:
            Begin
              Yy := Rclimite.Top;
              Y := Trunc(Yy);
              IF Abs(Vy) < 0.0001 Then
                Vy := 0
              Else
                Vy := -Vy * AmTop;
            End;
          2:
            Begin
              SetZero;
            End;
        End;
      End
    Else
      IF X > Rclimite.Right Then
        Begin
          ReBrect.Right := 1;
          Case Rebond.Right OF
            0:
              Begin
                Xx := Rclimite.Left;
                X := Trunc(Xx);
              End;
            1:
              Begin
                Xx := Rclimite.Right;
                X := Trunc(Xx);
                IF Abs(Vx) < 0.0001 Then
                  Vx := 0
                Else
                  Vx := -Vx * AmRight;
              End;
            2:
              Begin
                SetZero;
              End;
          End;
        End
      Else
        IF Y > Rclimite.Bottom Then
          Begin
            ReBrect.Bottom := 1;
            Case Rebond.Bottom OF
              0:
                Begin
                  Yy := Rclimite.Top;
                  Y := Trunc(Yy);
                End;
              1:
                Begin
                  Yy := Rclimite.Bottom;
                  Y := Trunc(Yy);
                  IF Abs(Vy) < 0.0001 Then
                    Vy := 0
                  Else
                    Vy := -Vy * AmBottom;
                End;
              2:
                Begin
                  SetZero;
                End;
            End;
          End;
End;

{ ======================================================================== }
CONSTRUCTOR TPool.Create(MaxS: Integer);
Var
  I: Integer;
Begin
  INHERITED Create;
  SpriteFirst := Nil;
  SpriteLast := Nil;
  PoolState := 0;
  Palbum := Nil;
  Maxspr := MaxS;
  Tableau := Tlist.Create;
  Tableau.Capacity := MaxS + 1;
  FOR I := 0 TO Maxspr DO
    Tableau.Add(Nil);
  Style := 0;
End;

Procedure TPool.Free;
Var
  I: Integer;
  S: TSprite;
  D: TToon;
Begin
  FOR I := 0 TO Maxspr DO
    Begin
      S := GetSprite(I);
      IF Assigned(S) Then
        Begin
          D := S.Toon;
          IF Assigned(D) Then
            Begin
              S.RcModif := Bounds(S.X, S.Y, D.DWidth, D.DHeight);
              With TAlbum(Palbum) Do
                Begin
                  BmpMano.Canvas.CopyMode := Cmsrccopy;
                  BmpMano.Canvas.CopyRect(S.RcModif, BmpFond.Canvas, S.RcModif);
                  TAlbum(Palbum).Cancan.CopyMode := Cmsrccopy;
                  TAlbum(Palbum).Cancan.CopyRect(S.RcModif, BmpMano.Canvas, S.RcModif);
                End;
            End;
        End;
      IF Assigned(Tableau[I]) Then
        TSprite(Tableau.Items[I]).Free;
    END;
  Tableau.Free;
  INHERITED Free;
  // Self := Nil;
End;

Function TPool.SpriteCount: Integer;
Var
  I, J: Integer;
Begin
  // count sprites
  J := 0;
  For I := 0 To Maxspr DO
    IF Assigned(Tableau.Items[I]) Then
      Inc(J);
  Result := J;
End;

Function TPool.GetSprite(N: Integer): TSprite;
Begin
  Result := Nil;
  IF (N < 0) OR (N > Maxspr) Then
    Exit;
  IF Assigned(Tableau.Items[N]) Then
    Result := TSprite(Tableau.Items[N]);
End;

Procedure TPool.Chainlast(S: TSprite);
Var
  Prec: TSprite;
Begin
  If S = Nil Then
    Exit;
  If SpriteLast = Nil Then
    Begin
      S.Next := Nil;
      S.Prior := Nil;
      SpriteFirst := S;
      SpriteLast := S;
    End
  Else
    Begin
      Prec := SpriteLast;
      S.Next := Nil;
      S.Prior := Prec;
      Prec.Next := S;
      SpriteLast := S;
    End;
End;

Procedure TPool.ChainSprite(S: TSprite);
Var
  Cur, Prec: TSprite;
  Ok: Boolean;
BEGIN
  IF S = Nil Then // no sprite
    Exit;
  IF SpriteFirst = Nil THEN // single sprite
    BEGIN
      S.Next := Nil;
      S.Prior := Nil;
      SpriteFirst := S;
      SpriteLast := S;
    END
  ELSE
    BEGIN
      Cur := SpriteFirst;
      IF S.Priority <= Cur.Priority Then // first in chain
        Begin
          S.Next := Cur;
          S.Prior := Nil;
          Cur.Prior := S;
          SpriteFirst := S;
        End
      Else
        Begin
          Ok := False;
          Repeat
            Prec := Cur;
            Cur := Cur.Next;
            IF Assigned(Cur) Then
              IF S.Priority <= Cur.Priority Then
                Ok := True;
          Until (Cur = Nil) OR Ok;
          IF Cur = Nil Then
            Begin
              S.Next := Nil;
              S.Prior := Prec;
              Prec.Next := S;
              SpriteLast := S;
            End
          Else
            Begin
              S.Next := Cur;
              S.Prior := Prec;
              Prec.Next := S;
              Cur.Prior := S;
            End;
        End;
    END;
END;

Procedure TPool.DechainSprite(S: TSprite);
Begin
  IF S = Nil Then
    Exit;
  IF (S.Prior = Nil) And (S.Next = Nil) Then // single sprite
    Begin
      SpriteFirst := Nil;
      SpriteLast := Nil;
      Exit;
    End;
  IF S.Prior = Nil Then // first in chain
    Begin
      SpriteFirst := S.Next;
      SpriteFirst.Prior := Nil;
    End
  Else
    Begin
      IF S.Next = Nil Then // last in chain
        Begin
          SpriteLast := S.Prior;
          S.Prior.Next := Nil
        End
      Else
        Begin
          S.Prior.Next := S.Next; // in chain
          S.Next.Prior := S.Prior;
        End
    End;
  S.Prior := Nil; // sprite
  S.Next := Nil;
End;

Function TPool.CreateSprite(N: Integer; Toonname: String): Boolean;
Var
  D: TToon;
Begin
  D := TAlbum(Palbum).GetToon(Toonname);
  IF D = Nil Then
    Begin
      Result := False;
      Exit;
    End;
  Result := SpriteCreate(N, 0);
  IF Result Then
    SetToon(N, Toonname);
End;

Function TPool.SpriteCreate(N: Integer; SpriteType: Integer): Boolean;
Var
  S: TSprite;
Begin
  IF (N < 0) OR (N > Maxspr) Then
    Begin
      Result := False;
      Exit;
    End;
  S := TSprite.Create;
  IF S = Nil Then
    Begin
      Result := False;
      Exit;
    End;
  IF Assigned(Tableau.Items[N]) Then
    SpriteDestroy(N);
  Tableau.Items[N] := S;
  ChainSprite(S);
  Result := True;
  S.Rcbounds := Rect(0, 0, TAlbum(Palbum).Fondwidth, TAlbum(Palbum).Fondheight);
  S.Rclimite := S.Rcbounds;
  S.Numero := N;
End;

Procedure TPool.SpriteErase(N: Integer);
Var
  S: TSprite;
  D: TToon;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      D := S.Toon;
      SetBit(Visible, 0, S.SprState);
      IF Assigned(D) Then
        Begin
          S.RcModif := Bounds(S.X, S.Y, D.DWidth, D.DHeight);
          With TAlbum(Palbum) Do
            Begin
              BmpMano.Canvas.CopyMode := Cmsrccopy;
              BmpMano.Canvas.CopyRect(S.RcModif, BmpFond.Canvas, S.RcModif);
              TAlbum(Palbum).Cancan.CopyMode := Cmsrccopy;
              TAlbum(Palbum).Cancan.CopyRect(S.RcModif, BmpMano.Canvas, S.RcModif);
            End;
        End;
    End;
End;

Procedure TPool.SpriteDestroy(N: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF S = Nil Then
    Exit;
  IF Getbit(Trace, S.SprState) = False Then
    SpriteErase(N);
  DechainSprite(S);
  S.Free;
  Tableau.Items[N] := Nil;
End;

Function TPool.SpriteExists(N: Integer): Boolean;
Begin
  Result := Assigned(GetSprite(N))
End;

Procedure TPool.SetTrace(N: Integer; B: Boolean);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF S = Nil Then
    Exit;
  IF B Then
    SetBit(Trace, 1, S.SprState)
  Else
    SetBit(Trace, 0, S.SprState);
End;

Procedure TPool.SetToon(N: Integer; Wname: STRING);
Var
  D: TToon;
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      D := TAlbum(Palbum).GetToon(Wname);
      IF D = Nil Then
        Begin
          S.Toon := Nil;
          SetBit(Visible, 0, S.SprState);
          Exit;
        End;
      SetBit(Visible, 1, S.SprState);
      S.Toon := D;
      S.CurImage := 0;
      S.Dcx := D.DWidth DIV 2;
      S.Dcy := D.DHeight DIV 2;
      S.Adjustlimits;
      SetBit(Modif, 1, S.SprState);
    End;
End;

Function TPool.GetToonName(N: Integer): STRING;
Var
  S: TSprite;
Begin
  Result := '';
  S := GetSprite(N);
  IF Assigned(S) Then
    IF Assigned(S.Toon) Then
      Result := TAlbum(Palbum).GetAlbumToonName(S.Toon);
End;

Function TPool.GetToonW(N: Integer): Integer;
Var
  S: TSprite;
Begin
  Result := 0;
  S := GetSprite(N);
  IF Assigned(S) Then
    IF Assigned(S.Toon) Then
      Result := S.Toon.DWidth;
End;

Function TPool.GetToonH(N: Integer): Integer;
Var
  S: TSprite;
Begin
  Result := 0;
  S := GetSprite(N);
  IF Assigned(S) Then
    IF Assigned(S.Toon) Then
      Result := S.Toon.DHeight;
End;

Procedure TPool.SetLoopCtr(N, M: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Loopmax := Abs(M);
      S.LoopCtr := S.Loopmax;
    End;
End;

Function TPool.GetLoopCtr(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.LoopCtr
  Else
    Result := 0;
End;

Procedure TPool.SetDelayCtr(N, M: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    S.DelayCtr := Abs(M);
End;

Function TPool.GetDelayCtr(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.DelayCtr
  Else
    Result := 0;
End;

Function TPool.PtinSprite(Pt: Tpoint; Marge: Integer): Integer;
Var
  S: TSprite;
Begin
  Result := -1;
  If SpriteLast = Nil Then
    Exit
  Else
    S := SpriteLast;
  Repeat
    If (Pt.X > S.X + Marge) And (Pt.Y > S.Y + Marge) And (Pt.X + Marge < S.X + S.Toon.DWidth) And
      (Pt.Y + Marge < S.Y + S.Toon.DHeight) Then
      Begin
        Result := S.Numero;
        Exit;
      End;
    S := S.Prior;
  Until S = Nil;
End;

Procedure TPool.PoolRemove; // efface
Var
  S: TSprite;
  D: TToon;
Begin
  If SpriteFirst = Nil Then
    Exit
  Else
    S := SpriteFirst;
  Repeat
    IF Getbit(Actif, S.SprState) Then
      Begin
        D := S.Toon;
        IF Assigned(D) Then
          Begin
            S.RcModif := Bounds(S.X, S.Y, S.RcImg.Right, S.RcImg.Bottom);
            IF Getbit(Trace, S.SprState) = False Then
              Begin
                With TAlbum(Palbum) Do
                  Begin
                    BmpMano.Canvas.CopyMode := Cmsrccopy;
                    BmpMano.Canvas.CopyRect(S.RcModif, BmpFond.Canvas, S.RcModif);
                  End;
              End;
          End;
      End;
    S := S.Next;
  Until S = Nil;
End;

Procedure TPool.PoolMove; // deplace
Var
  S: TSprite;
  D: TToon;
Begin
  If SpriteFirst = Nil Then
    Exit
  Else
    S := SpriteFirst;
  Repeat
    D := S.Toon;
    IF Assigned(D) AND Getbit(Modif, S.SprState) Then
      Begin
        S.RcImg.Right := S.Toon.DWidth;
        S.RcImg.Bottom := S.Toon.DHeight;
        SetBit(Modif, 0, S.SprState);
      End;
    IF Getbit(Actif, S.SprState) Then
      Begin
        S.Calculpos;
        IF S.Loopmax <> 0 Then
          Begin
            Dec(S.LoopCtr);
            IF S.LoopCtr < 0 Then
              S.LoopCtr := S.Loopmax;
          End;
      End;
    IF S.DelayCtr > 0 Then
      Dec(S.DelayCtr);
    S := S.Next;
  Until S = Nil;
End;

Procedure TPool.PrepareCache;
Var
  S: TSprite;
  D: TToon;
Begin
  If SpriteFirst = Nil Then
    Exit
  Else
    S := SpriteFirst;
  Repeat
    IF Getbit(Actif, S.SprState) Then
      Begin
        D := S.Toon;
        IF Assigned(D) AND Getbit(Visible, S.SprState) Then
          Begin
            S.RcMano := Bounds(S.X, S.Y, D.DWidth, D.DHeight);
            WITH TAlbum(Palbum).BmpMano.Canvas DO
              Begin
                CopyMode := CmSrcAnd;
                CopyRect(S.RcMano, Tbitmap(D.Masks.Items[S.CurImage]).Canvas, S.RcImg);
                CopyMode := CmSrcpaint;
                CopyRect(S.RcMano, Tbitmap(D.Images.Items[S.CurImage]).Canvas, S.RcImg);
              End;
          End;
      End;
    S := S.Next;
  Until S = Nil;
End;

Procedure TPool.SpritesShow;
Var
  S: TSprite;
  D: TToon;
Begin
  TAlbum(Palbum).Cancan.CopyMode := Cmsrccopy;
  If SpriteFirst = Nil Then
    Exit
  Else
    S := SpriteFirst;
  Log('SpriteShow');
  Repeat
    IF Getbit(Actif, S.SprState) Then
      Begin
        D := S.Toon;
        IF Assigned(D) And Getbit(Visible, S.SprState) Then
          Begin
            UnionRect(S.RcModif, S.RcModif, S.RcMano);
            With TAlbum(Palbum) Do
              Begin
                Cancan.CopyRect(S.RcModif, BmpMano.Canvas, S.RcModif);
              End;
          End;
      End;
    S := S.Next;
  Until S = Nil;
End;

Procedure TPool.Animate;
Begin
  IF Not Albumdraw Then
    Begin
      PoolRemove; // effac
      PoolMove; // decpla
      PrepareCache;
      SpritesShow;
    End;
End;

Procedure TPool.AnimateMono(N: Integer);
Var
  Sn, S: TSprite;
  D, Dn: TToon;
  Collisrect: Trect;
Begin
  S := GetSprite(N);
  Log('& Animate Mono '+ inttostr(N)); // Single Part
  IF Assigned(S) Then
    Begin
      D := S.Toon;
      IF Assigned(D) Then // in album
        Begin
          If Getbit(Modif, S.SprState) Then
            S.RcModif := Bounds(S.X, S.Y, S.RcImg.Right, S.RcImg.Bottom)
          Else
            S.RcModif := Bounds(S.X, S.Y, D.DWidth, D.DHeight);

          With TAlbum(Palbum) Do
            Begin
              BmpMano.Canvas.CopyMode := Cmsrccopy;
              BmpMano.Canvas.CopyRect(S.RcModif, BmpFond.Canvas, S.RcModif);
            End;

          Sn := SpriteFirst;
          Repeat
            If (Sn <> S) And Getbit(Visible, Sn.SprState) Then
              Begin
                Dn := Sn.Toon;
                Sn.RcModif := Bounds(Sn.X, Sn.Y, Dn.DWidth, Dn.DHeight);
{$IFDEF Win32}
                IF IntersectRect(Collisrect, Sn.RcModif, S.RcModif) Then
{$ELSE}
  {$IFDEF Win64}
                IF IntersectRect(Collisrect, Sn.RcModif, S.RcModif) Then
  {$ELSE}
                IF IntersectRect(Collisrect, Sn.RcModif, S.RcModif) <> 0 Then
  {$IFEND}
{$IFEND}
                  Begin
                    Sn.RcMano := Bounds(Sn.X, Sn.Y, Dn.DWidth, Dn.DHeight);
                    WITH TAlbum(Palbum).BmpMano.Canvas DO
                      Begin
                        CopyMode := CmSrcAnd;
                        CopyRect(Sn.RcMano, Tbitmap(Dn.Masks.Items[Sn.CurImage]).Canvas, Sn.RcImg);
                        CopyMode := CmSrcpaint;
                        CopyRect(Sn.RcMano, Tbitmap(Dn.Images.Items[Sn.CurImage]).Canvas, Sn.RcImg);
                      End;
                  End;
              End;
            Sn := Sn.Next;
          Until Sn = Nil;
          IF Assigned(D) AND Getbit(Modif, S.SprState) Then
            Begin
              S.RcImg.Right := S.Toon.DWidth;
              S.RcImg.Bottom := S.Toon.DHeight;
              SetBit(Modif, 0, S.SprState);
            End;
          S.Calculpos;
          IF S.Loopmax <> 0 Then
            Begin
              Dec(S.LoopCtr);
              IF S.LoopCtr < 0 Then
                S.LoopCtr := S.Loopmax;
            End;
          IF S.DelayCtr > 0 Then
            Dec(S.DelayCtr);

          S.RcMano := Bounds(S.X, S.Y, D.DWidth, D.DHeight);
          WITH TAlbum(Palbum).BmpMano.Canvas DO
            Begin
              CopyMode := CmSrcAnd;
              CopyRect(S.RcMano, Tbitmap(D.Masks.Items[S.CurImage]).Canvas, S.RcImg);
              CopyMode := CmSrcpaint;
              CopyRect(S.RcMano, Tbitmap(D.Images.Items[S.CurImage]).Canvas, S.RcImg);
            End;
          TAlbum(Palbum).Cancan.CopyMode := Cmsrccopy;
          UnionRect(S.RcModif, S.RcModif, S.RcMano);
          With TAlbum(Palbum) Do
            Begin
              Cancan.CopyRect(S.RcModif, BmpMano.Canvas, S.RcModif);
            End;
        End;
    End;
End;

Procedure TPool.SetLimits(N: Integer; Wrect: Trect);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Rcbounds := Wrect;
      S.Adjustlimits;
    End;
End;

Procedure TPool.SetActive(N: Integer; B: Boolean);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    IF B Then
      SetBit(Actif, 1, S.SprState)
    Else
      SetBit(Actif, 0, S.SprState);
End;

Function TPool.IsActive(N: Integer): Boolean;
Var
  S: TSprite;
Begin
  Result := False;
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := Getbit(Actif, S.SprState);
End;

Procedure TPool.SetVisible(N: Integer; B: Boolean);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    IF B Then
      SetBit(Visible, 1, S.SprState)
    Else
      SetBit(Visible, 0, S.SprState);
End;

Function TPool.IsVisible(N: Integer): Boolean;
Var
  S: TSprite;
Begin
  Result := False;
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := Getbit(Actif, S.SprState);
End;

Procedure TPool.Settag(N, Atag: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    S.Sprtag := Atag;
End;

Function TPool.Gettag(N: Integer): Integer;
Var
  S: TSprite;
Begin
  Result := 0;
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Sprtag;
End;

Procedure TPool.SetAnim(N: Integer; Wcount: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    S.Animloop := Wcount;
End;

Procedure TPool.SetAnimStyle(N, Wstyle: Integer);
Var
  S: TSprite;
  D: TToon;
  Bmp1: Tbitmap;

  Procedure Rot90; // rotate 90 degrees
  Type
    TManoRGB = Packed Record
      Rgb: TRGBTriple;
      Dummy: Byte;
    End;

    TRGBArray = ARRAY [0 .. 0] OF TRGBTriple;
    PRGBArray = ^TRGBArray;
  Var
    AStream: TMemorystream;
    Header: TBITMAPINFO;
    Dc: HDC;
    P: ^TManoRGB;
    RowOut: PRGBArray;
    X, Y, H, W: Integer;
  BEGIN
    IF Bmp1.Empty Then
      Exit;
    Bmp1.Pixelformat := Pf24bit;
    AStream := TMemorystream.Create;
    AStream.SetSize(Bmp1.Height * Bmp1.Width * 4);
    With Header.BmiHeader Do
      Begin
        BiSize := SizeOf(TBITMAPINFOHEADER);
        BiWidth := Bmp1.Width;
        BiHeight := Bmp1.Height;
        BiPlanes := 1;
        BiBitCount := 32;
        BiCompression := 0;
        BiSizeimage := AStream.Size;
        BiXPelsPerMeter := 1;
        BiYPelsPerMeter := 1;
        BiClrUsed := 0;
        BiClrImportant := 0;
      End;
    Dc := GetDC(0);
    P := AStream.Memory;

    GetDIBits(Dc, Bmp1.Handle, 0, Bmp1.Height, P, Header, Dib_RGB_Colors);
    ReleaseDC(0, Dc);
    W := Bmp1.Height; // swap sides
    H := Bmp1.Width;
    Bmp1.Width := W;
    Bmp1.Height := H;
    For Y := 0 To (H - 1) Do
      Begin
        RowOut := Bmp1.ScanLine[Y];
        P := AStream.Memory;
        Inc(P, Y);
        For X := 0 To (W - 1) Do
          Begin
            RowOut[X] := P^.Rgb;
            Inc(P, H);
          End;
      End;
    AStream.Free;
  End;

Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      IF Wstyle > 3 Then
        Wstyle := 3;
      If Wstyle < 0 Then
        Wstyle := 0;
      D := S.Toon;
      IF Assigned(D) Then
        Begin
          Bmp1 := Tbitmap(D.Images.Items[0]);
          Repeat
            Inc(S.Animstyle); // rotate 90 degrees clockwise
            If S.Animstyle > 3 Then
              S.Animstyle := 0;
            Rot90;
          Until S.Animstyle = Wstyle;
          D.DWidth := Bmp1.Width;
          D.DHeight := Bmp1.Height;
          Tbitmap(D.Masks.Items[0]).Width := D.DWidth;
          Tbitmap(D.Masks.Items[0]).Height := D.DHeight;
          Domask(Bmp1, Tbitmap(D.Masks.Items[0]), D.DWidth, D.DHeight, Clblack);
          SetBit(Modif, 1, S.SprState);
        End;
    End;
End;

Function TPool.GetCurimage(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Animstyle
  Else
    Result := 0;
End;

Procedure TPool.Setpriority(N, P: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Priority := P;
      DechainSprite(S);
      ChainSprite(S);
    End;
End;

Procedure TPool.Setprioritylast(N, P: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Priority := P;
      DechainSprite(S);
      Chainlast(S);
    End;
End;

Function TPool.Getpriority(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Priority
  Else
    Result := 0;
End;

Procedure TPool.SetPos(N, Wx, Wy: Integer); // set to position Wx,Wy
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Xx := Trunc(Wx);
      S.Yy := Trunc(Wy);
    End;
End;

Function TPool.Getposx(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := Trunc(S.Xx)
  Else
    Result := 0;
End;

Function TPool.Getposy(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := Trunc(S.Yy)
  Else
    Result := 0;
End;

Function TPool.Getdcx(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Dcx
  Else
    Result := 0;
End;

Function TPool.Getdcy(N: Integer): Integer;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Dcy
  Else
    Result := 0;
End;

Procedure TPool.SetV(N: Integer; Wvx, Wvy: Single); // Move  Wvx,Wvy
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Vx := Wvx;
      S.Vy := Wvy;
    End;
End;

Function TPool.Getvx(N: Integer): Single;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Vx
  Else
    Result := 0;
End;

Function TPool.Getvy(N: Integer): Single;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Vy
  Else
    Result := 0;
End;

Function TPool.SameXdirection(N: Integer): Boolean;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  Result := True;
  IF Assigned(S) Then
    IF S.Vx * S.Oldvx < 0 Then
      Result := False;
End;

Function TPool.SameYdirection(N: Integer): Boolean;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  Result := True;
  IF Assigned(S) Then
    IF S.Vy * S.Oldvy < 0 Then
      Result := False;
End;

Procedure TPool.Setaccel(N: Integer; Wax, Way: Single);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Ax := Wax;
      S.Ay := Way;
    End;
End;

Procedure TPool.SetKV(N: Integer; Kvx, Kvy: Single);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Kvx := Abs(Kvx);
      S.Kvy := Abs(Kvy);
    End;
End;

Procedure TPool.Stop(N: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Vx := 0;
      S.Vy := 0;
      S.Ax := 0;
      S.Ay := 0;
      S.Kvx := 1;
      S.Kvy := 1;
      S.Animloop := 0;
    End;
End;

Function TPool.Getax(N: Integer): Single;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Ax
  Else
    Result := 0;
End;

Function TPool.Getay(N: Integer): Single;
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Ay
  Else
    Result := 0;
End;

Function TPool.Getaxvnull(N, X1: Integer): Single;
Var
  S: TSprite;
  B, D: Single;
Begin
  S := GetSprite(N);
  Result := 0;
  IF S = Nil Then
    Exit;
  IF Abs(S.Vx) < 0.00001 Then
    Exit;
  D := X1 - S.Xx;
  D := D + S.Vx;
  If Abs(D) < 2 Then
    Exit;
  If (Abs(S.Vx) * 3) > Abs(D) Then
    Exit;
  Result := -(S.Vx * S.Vx) / (2 * D);
  B := (2 * D) / S.Vx;
  S.Loopmax := Round(B);
  S.LoopCtr := S.Loopmax;
End;

Function TPool.Getayvnull(N, Y1: Integer): Single;
Var
  S: TSprite;
  B, D: Single;
Begin
  S := GetSprite(N);
  Result := 0;
  IF S = Nil Then
    Exit;
  IF Abs(S.Vy) < 0.00001 Then
    Exit;
  D := Y1 - S.Yy;
  D := D + S.Vy;
  If Abs(D) < 2 Then
    Exit;
  If (Abs(S.Vy) * 3) > Abs(D) Then
    Exit;
  Result := -(S.Vy * S.Vy) / (2 * D);
  B := (2 * D) / S.Vy;
  S.Loopmax := Round(B);
  S.LoopCtr := S.Loopmax;
End;

Procedure TPool.Gotoxy(N, Wx, Wy, Loops: Integer);
Var
  S: TSprite;
  Wdx, Wdy, Wb: Single;
Begin
  S := GetSprite(N);
  IF (S = Nil) OR (Loops < 1) Then
    Exit;
  S.Kvx := 1;
  S.Kvy := 1;
  Wb := Loops;
  S.Ax := -S.Vx * 2 / Wb;
  S.Ay := -S.Vy * 2 / Wb;
  Wdx := Wx;
  Wdy := Wy;
  Wdx := Wdx - S.Xx;
  Wdy := Wdy - S.Yy;
  S.Vx := S.Vx + Wdx / Wb;
  S.Vy := S.Vy + Wdy / Wb;
End;

Function TPool.Getoffsetx(N: Integer): Integer;
Var
  S: TSprite;
Begin
  Result := 0;
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Toon.Origpt.X;
End;

Function TPool.Getoffsety(N: Integer): Integer;
Var
  S: TSprite;
Begin
  Result := 0;
  S := GetSprite(N);
  IF Assigned(S) Then
    Result := S.Toon.Origpt.Y;
End;

Procedure TPool.Setoffset(N, Wx, Wy: Integer);
Var
  S: TSprite;
Begin
  S := GetSprite(N);
  IF Assigned(S) Then
    Begin
      S.Toon.Origpt.X := Wx;
      S.Toon.Origpt.Y := Wy;
    End;
End;

Begin
  Initrandom;
End.
