Unit UIniOptions;

// store all settings in inifile

Interface

Uses
{$IF CompilerVersion > 28} // Delphi XE7
  Winapi.Windows, System.SysUtils, System.Classes, System.Types, System.UITypes, vcl.Forms, vcl.Graphics, System.IniFiles;
{$ELSE}
  Classes, SysUtils, IniFiles, Types, Forms, Windows, Graphics;
{$IFEND}

Const
  CsIniPuzzleSection = 'Puzzle';
  CsIniJigSawSection = 'JigSaw';
  CsIniPiecesIdent = 'Pieces';
  CsIniScreenIdent = 'Screen';
  CsIniTempIdent = 'Temps';
  CsIniPartIdent = 'Part';

  MagentLow = 4; // magnetic attraction range
  MagnetHigh = 12;

Type

  TNumArray = Array [1 .. 6] Of Integer; // decode 6 numeric values

  TIniOptions = Class(TObject)
  Private
    FSize: Integer;
    FSpeed: Integer;
    FNiveau: Integer;
    FPieceBorderColor: Integer;
    FRotatepieces: Boolean;
    FPicture: String;
    FMagnet: Integer;
    FUseCtrl: Boolean;
    FAutoSolve: Boolean;
    FStartBorderPieces: Boolean;
    FShowTime: Boolean;
    FIniFilename: String;
    FModified: Boolean;

    FOrgNiveau: Integer;
    FOrgColor: Integer;
    FOrgRotate: Boolean;
    FShowLog: Boolean;
    FPlace: TRect;
    FPreviewPlace: TRect;
    Function StrtoRect(Value: String): TRect;
    Function RecttoStr(Value: TRect): String;
    Procedure SetMagnet(Const Value: Integer);
    Procedure SetNiveau(Const Value: Integer);
    Procedure SetPicture(Const Value: String);
    Procedure SetPieceBorderColor(Const Value: Integer);
    Procedure SetRotatepieces(Const Value: Boolean);
    Procedure SetSpeed(Const Value: Integer);
    Procedure SetSize(Const Value: Integer);
    Procedure SetUseCtrl(Const Value: Boolean);
    Procedure SetAutoSolve(Const Value: Boolean);
    Procedure SetStartBorderPieces(Const Value: Boolean);
    Procedure SetShowTime(Const Value: Boolean);
    Procedure SetIniFilename(Const Value: String);
    Procedure SetModified(Const Value: Boolean);
    procedure SetShowLog(const Value: Boolean);
    procedure SetPlace(const Value: TRect);
    procedure SetPreviewPlace(const Value: TRect);
  Public
    Procedure LoadSettings(Ini: TIniFile);
    Procedure SaveSettings(Ini: TIniFile);

    Procedure LoadFromFile(Const FileName: String);
    Procedure SaveToFile(Const FileName: String);

    Function LoadIdentFromIniFile(Ident: String): TNumArray;
    Procedure SaveIdentToIniFile(Ident, Value: String);

    Procedure CleanUpIniFile;
    Procedure StoreOptions;
    Procedure ReStoreOptions;

    Property IniFilename: String Read FIniFilename Write SetIniFilename;
    Property Magnet: Integer Read FMagnet Write SetMagnet;
    Property Niveau: Integer Read FNiveau Write SetNiveau;
    Property Picture: String Read FPicture Write SetPicture;
    Property PieceBorderColor: Integer Read FPieceBorderColor Write SetPieceBorderColor;
    Property Rotatepieces: Boolean Read FRotatepieces Write SetRotatepieces;
    Property Speed: Integer Read FSpeed Write SetSpeed;
    Property Size: Integer Read FSize Write SetSize;
    Property UseCtrl: Boolean Read FUseCtrl Write SetUseCtrl;
    Property AutoSolve: Boolean Read FAutoSolve Write SetAutoSolve;
    Property Place: TRect read FPlace write SetPlace;
    Property PreviewPlace: TRect read FPreviewPlace write SetPreviewPlace;
    Property StartBorderPieces: Boolean Read FStartBorderPieces Write SetStartBorderPieces;
    Property ShowLog: Boolean read FShowLog write SetShowLog;
    Property ShowTime: Boolean Read FShowTime Write SetShowTime;
    Property Modified: Boolean Read FModified Write SetModified;
  End;

Var
  IniOptions: TIniOptions = Nil;

Implementation

Procedure TIniOptions.LoadSettings(Ini: TIniFile);
Begin
  If Assigned(Ini) Then
    Begin
      IniOptions.Picture := Ini.ReadString(CsIniPuzzleSection, 'Picture', '');
      IniOptions.Magnet := Ini.ReadInteger(CsIniPuzzleSection, 'Magnet', MagnetHigh);
      IniOptions.Niveau := Ini.ReadInteger(CsIniPuzzleSection, 'Niveau', 2);
      IniOptions.Picture := Ini.ReadString(CsIniPuzzleSection, 'Picture', '');
      IniOptions.PieceBorderColor := Ini.ReadInteger(CsIniPuzzleSection, 'BorderColor', ClGray);
      IniOptions.Rotatepieces := Ini.ReadBool(CsIniPuzzleSection, 'Rotatepieces', True);
      IniOptions.Speed := Ini.ReadInteger(CsIniPuzzleSection, 'Speed', 2); // 0 = slow, 1 = middle, 2 = fast, 3 = very fast
      IniOptions.Size := Ini.ReadInteger(CsIniPuzzleSection, 'Size', 2); // 0 = real size, 1 = thumbnail, 2 = hide
      IniOptions.UseCtrl := Ini.ReadBool(CsIniPuzzleSection, 'UseCtrl', False);
      IniOptions.AutoSolve := Ini.ReadBool(CsIniPuzzleSection, 'AutoSolve', False);
      IniOptions.Place := StrtoRect(Ini.ReadString(CsIniPuzzleSection, 'Place', '10,10,100,100')); // minimum constraints
      IniOptions.PreviewPlace := StrtoRect(Ini.ReadString(CsIniPuzzleSection, 'Preview', '10,10,100,100'));
      IniOptions.StartBorderPieces := Ini.ReadBool(CsIniPuzzleSection, 'BorderPieces', False);
      IniOptions.ShowLog := Ini.ReadBool(CsIniPuzzleSection, 'ShowLog', False);
      IniOptions.ShowTime := Ini.ReadBool(CsIniPuzzleSection, 'ShowTime', True);
      IniOptions.FModified := False; // readonly property

    End;
End;

Procedure TIniOptions.SaveSettings(Ini: TIniFile);
Begin
  If Assigned(Ini) Then
    Begin
      Ini.WriteString(CsIniPuzzleSection, 'Picture', IniOptions.Picture);
      Ini.WriteInteger(CsIniPuzzleSection, 'Magnet', IniOptions.Magnet);
      Ini.WriteInteger(CsIniPuzzleSection, 'Niveau', IniOptions.Niveau);
      Ini.WriteString(CsIniPuzzleSection, 'Picture', IniOptions.Picture);
      Ini.WriteInteger(CsIniPuzzleSection, 'BorderColor', IniOptions.PieceBorderColor);
      Ini.WriteBool(CsIniPuzzleSection, 'Rotatepieces', IniOptions.Rotatepieces);
      Ini.WriteInteger(CsIniPuzzleSection, 'Speed', IniOptions.Speed);
      Ini.WriteInteger(CsIniPuzzleSection, 'Size', IniOptions.Size);
      Ini.WriteBool(CsIniPuzzleSection, 'UseCtrl', IniOptions.UseCtrl);
      Ini.WriteBool(CsIniPuzzleSection, 'AutoSolve', IniOptions.AutoSolve);
      Ini.WriteString(CsIniPuzzleSection, 'Place', RecttoStr(IniOptions.Place));
      Ini.WriteString(CsIniPuzzleSection, 'Preview', RecttoStr(IniOptions.PreviewPlace));
      Ini.WriteBool(CsIniPuzzleSection, 'ShowLog', IniOptions.ShowLog);
      Ini.WriteBool(CsIniPuzzleSection, 'ShowTime', IniOptions.ShowTime);
    End;
End;

// ----------------------------------------------------------------------------------------------
Procedure TIniOptions.CleanUpIniFile;

Var
  Ini: TIniFile;
Begin
  IniOptions.Picture := ''; // no picture
  Ini := TIniFile.Create(IniFilename);
  Try
    Ini.EraseSection(CsIniJigSawSection); // remove prev stored section
  Finally
    Ini.Free;
  End;
End;

Procedure TIniOptions.LoadFromFile(Const FileName: String);
Var
  Ini: TIniFile;
Begin
  Ini := TIniFile.Create(FileName);
  Try
    LoadSettings(Ini);
  Finally
    Ini.Free;
  End;
End;

Procedure TIniOptions.SaveToFile(Const FileName: String);
Var
  Ini: TIniFile;
Begin
  Ini := TIniFile.Create(FileName);
  Try
    SaveSettings(Ini);
  Finally
    Ini.Free;
  End;
End;

// ------------------------------------------------------------------------------------------------

Function TIniOptions.LoadIdentFromIniFile(Ident: String): TNumArray;
Var
  Ini: TIniFile;
  TempLine: String;

  Function Number6(Value: String): TNumArray; // value into 6 integers num[]
  Var
    P, I: Integer;
    TempLine: String;

    Function ClearResult: TNumArray; // cleararray
    Var
      I: Integer;
    Begin
      For I := 1 To 6 Do
        Result[I] := 0;
    End;

  Begin
    Result := ClearResult;
    I := 0;
    TempLine := Value;

    If TempLine <> '' Then // no value
      Begin
        Repeat
          P := Pos(';', TempLine);
          If P > 0 Then
            Begin
              Inc(I);
              Try
                Result[I] := Strtoint(Copy(TempLine, 1, P - 1));
              Except
              End;
              Delete(TempLine, 1, P);
            End;
        Until P = 0;
      End;
  End;

Begin
  Ini := TIniFile.Create(IniFilename);
  Try
    TempLine := Ini.ReadString(CsIniJigSawSection, Ident, '');
    Result := Number6(TempLine);
  Finally
    Ini.Free;
  End;
End;

// ------------------------------------------------------------------------------------------------

Procedure TIniOptions.SaveIdentToIniFile(Ident, Value: String);
Var
  Ini: TIniFile;
Begin
  Ini := TIniFile.Create(IniFilename);
  Try
    Ini.WriteString(CsIniJigSawSection, Ident, Value);
  Finally
    Ini.Free;
  End;
End;

// ------------------------------------------------------------------------------------------------

function TIniOptions.RecttoStr(Value: TRect): String;
begin
  Result := inttostr(Value.Left) + ',' + inttostr(Value.Top) + ',' + inttostr(Value.Right) + ',' + inttostr(Value.Bottom);
end;

Procedure TIniOptions.ReStoreOptions;
Begin
  FNiveau := FOrgNiveau;
  FRotatepieces := FOrgRotate;
  FPieceBorderColor := FOrgColor;
End;

Procedure TIniOptions.SetPicture(Const Value: String);
Begin
  FPicture := Value;
End;

Procedure TIniOptions.SetAutoSolve(Const Value: Boolean);
Begin
  FAutoSolve := Value;
End;

Procedure TIniOptions.SetIniFilename(Const Value: String);
Begin
  FIniFilename := Value;
End;

Procedure TIniOptions.SetMagnet(Const Value: Integer);
Begin
  FMagnet := Value;
End;

Procedure TIniOptions.SetModified(Const Value: Boolean);
Begin
  FModified := Value;
End;

Procedure TIniOptions.SetNiveau(Const Value: Integer);
Begin
  If FNiveau <> Value Then
    Begin
      FNiveau := Value;
      FModified := True;
    End;
End;

Procedure TIniOptions.SetPieceBorderColor(Const Value: Integer);
Begin
  FPieceBorderColor := Value;
End;

procedure TIniOptions.SetPlace(const Value: TRect);
begin
  FPlace := Value;
end;

procedure TIniOptions.SetPreviewPlace(const Value: TRect);
begin
  FPreviewPlace := Value;
end;

Procedure TIniOptions.SetRotatepieces(Const Value: Boolean);
Begin
  If FRotatepieces <> Value Then
    Begin
      FRotatepieces := Value;
      FModified := True;
    End;
End;

procedure TIniOptions.SetShowLog(const Value: Boolean);
begin
  FShowLog := Value;
end;

Procedure TIniOptions.SetShowTime(Const Value: Boolean);
Begin
  FShowTime := Value;
End;

Procedure TIniOptions.SetSize(Const Value: Integer);
Begin
  If FSize <> Value Then
    Begin
      FSize := Value;
      FModified := True;
    End;
End;

Procedure TIniOptions.SetSpeed(Const Value: Integer);
Begin
  FSpeed := Value;
End;

Procedure TIniOptions.SetStartBorderPieces(Const Value: Boolean);
Begin
  FStartBorderPieces := Value;
End;

Procedure TIniOptions.SetUseCtrl(Const Value: Boolean);
Begin
  FUseCtrl := Value;
End;

Procedure TIniOptions.StoreOptions;
Begin
  FOrgNiveau := FNiveau; // use settings at FormGameoptions
  FOrgRotate := FRotatepieces;
  FOrgColor := FPieceBorderColor;
End;

function TIniOptions.StrtoRect(Value: String): TRect;
var
  templist: Tstringlist;
begin
  if Value <> '' then
    begin
      templist := Tstringlist.Create;
      Result.Empty;
      try
        templist.commatext := Value;
        Result.Left := Strtoint(templist[0]);
        Result.Top := Strtoint(templist[1]);
        Result.Right := Strtoint(templist[2]);
        Result.Bottom := Strtoint(templist[3]);
      finally
        templist.Free;
      end;
    end;
end;

Initialization

IniOptions := TIniOptions.Create;
IniOptions.IniFilename := ChangeFileExt(Application.ExeName, '.Ini');
IniOptions.LoadFromFile(IniOptions.IniFilename);

Finalization

IniOptions.Free;

End.
