Unit UMain;

// Orginal Project Qpuzzle - (c) Jean-Yves Quinec 2000
// Rewritten and modifed for use Delphi2010 - Delphi 10.3
Interface

Uses
{$IF CompilerVersion > 28} // Delphi XE7
  Winapi.Windows, Winapi.Messages, Winapi.CommCtrl, System.SysUtils, System.Classes, Vcl.Graphics, Vcl.Controls, Vcl.Forms,
  Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons, Vcl.ExtCtrls, System.Types, System.UITypes, Vcl.ExtDlgs, Vcl.Menus, Vcl.ComCtrls,
{$ELSE}
  Windows, Messages, SysUtils, Classes, Forms, Graphics, Dialogs, ExtDlgs, Menus,
  ExtCtrls, ComCtrls, CommCtrl, Controls, StdCtrls,
{$IFEND}
  Jpeg, Spritez, HtmlHelpviewer, ImgList;

Const
  MaxPieces = 128; // maximum pieces (rows and columns)
  BezierLineSteps = 9; // line with bezier

Type
  Tpiece = Record
    Pleft: Byte; // the 4 sides are coded:
    Ptop: Byte; // 0 border, 1 kader externe, 2 kader interne
    Pright: Byte;
    Pbottom: Byte;
    Pinit: Tpoint; // initial piece coordinates (redo)
    PRotation: Integer; // Current rotation 0 .. 3
    PRotation0: Integer; // initial rotation (redo)
    Pflag: Integer; // pieces state 1 = placed in final position
  End;

  TPaintFlag = (PfEmpty, PfImage, PfPart);

  TFormMain = Class(TForm)
    TimerTime: TTimer;
    StatusBar: TStatusBar;
    ProgressBar: TProgressBar;
    Paintbox: TPaintBox;
    LabelSolved: TLabel;
    OpenPictureDialog: TOpenPictureDialog;

    MainMenu: TMainMenu;
    MenuExit: TMenuItem;
    MenuNewImageOud: TMenuItem;
    MenuOptions: TMenuItem;
    MenuShowImage: TMenuItem;
    MenuSolution: TMenuItem;
    MenuHelp: TMenuItem;
    MenuGame: TMenuItem;
    N2: TMenuItem;
    MenuAbout: TMenuItem;
    MenuStartAgain: TMenuItem;
    MenuHelpContens: TMenuItem;
    MenuRotatePieces: TMenuItem;
    MenuBorderColor: TMenuItem;
    MemoLog: TMemo;
    Splitter1: TSplitter;
    PanelPuzzle: TPanel;
    NiveauGroup: TMenuItem;
    MenuNiveau4: TMenuItem;
    MenuNiveau3: TMenuItem;
    MenuNiveau2: TMenuItem;
    MenuNiveau1: TMenuItem;
    MenuSettings: TMenuItem;
    PopupMenu1: TPopupMenu;

    Procedure FormActivate(Sender: TObject);
    Procedure FormCreate(Sender: TObject);
    Procedure FormClose(Sender: TObject; Var Action: TCloseAction);
    Procedure FormDestroy(Sender: TObject);
    Procedure PaintboxPaint(Sender: TObject);
    Procedure PaintboxMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Procedure PaintboxMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    Procedure PaintboxMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Procedure TimerTimeTimer(Sender: TObject);
    Procedure StatusBarResize(Sender: TObject);
    Procedure MenuShowImageClick(Sender: TObject);
    Procedure MenuSolutionClick(Sender: TObject);
    Procedure MenuNewImageClick(Sender: TObject);
    Procedure MenuHelpAboutClick(Sender: TObject);
    Procedure MenuNewStartClick(Sender: TObject);
    Procedure MenuAboutExecute(Sender: TObject);
    Procedure MenuHelpExecute(Sender: TObject);
    Procedure MenuExitClick(Sender: TObject);
    Procedure MenuNiveauClick(Sender: TObject);
    Procedure MenuRotatePiecesClick(Sender: TObject);
    Procedure MenuBorderColorClick(Sender: TObject);
    Procedure MenuSetting;
    Procedure MenuSettingsClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
  Private
    Startdir: String;
    PaintboxFlag: TPaintFlag; // 0 empty, 1 draws Bmp1, 2 draw pieces

    Completed: Boolean;
    ImageIsLoaded: Boolean; // true when image is loaded
    SpritesCreated: Boolean; // true when sprites are created
    MoveAuto: Boolean; // true when a piece in auto displacement state
    Moving: Boolean; // true when mouse displacement
    Redo: Boolean; // true when redo puzzle

    PieceNr: Integer; // Current sprite number
    PieceRow: Integer; // Current puzzle line
    PieceCol: Integer; // Current pueele column
    Mx, My: Integer; // memorise mouse coordinates
    Speed: Integer; // timer loop speed
    // stats
    StartTime: Tdatetime; // start time
    Timebefore: Tdatetime; // prior game duration
    CountAutoMove: Integer; // Auto pieces move counter
    CountMouseMove: Integer; // mouse Pieces move counter

    // Parts Design
    // Virtual       k3 :         :     k2
    // origin     >+----+---------+--------------+---------
    // px,py       :    :         :              :
    //             :    : origin  :              :
    //             +----+---------+--------------+------       Piece size [0,0]
    //             :    |       __|              |             width  k2-k3
    //             :    |      /                 |             height k2-k3+k3 = k2
    //             :    |[0,0] \__:--------------|---:         x0 = px+k3, y0 = py+k3
    //             :    |         |      /\      |   :
    //             :    |         |   k3|  |     |   :         Piece size  [1,0]
    // px,py       +----+--,  ,---+-----'  '-----+---:-   width  k2-k3
    //             :    |  |  |   |px,py         |   :         height k2+k3
    //             :    |   \/    |              |<k3>         x0 = px+k3, y0 =py
    //             :    |         |__            |__ :
    // k2          :    |[1,0]       \ [1,1]        \:         Piece size  [1,1]
    //             :    |          __/            __/:         width  k2+k3
    //             :    |         |              |   :         height k3+k2+k3
    //             :    |         |              |   :         x0 = px, y0 = py-k3
    //             :    |         |              |   :
    //             :----+--,  ,---+-----,  ,-----+---+---------------
    //             :    |  |  |   |     |  |     |   :
    //             :k3  |   \/    :______\/__________:

    Bmp1: Tbitmap; // stretched image bitmap to hafe an exact numner of pieces
    PartsArray: Array [0 .. MaxPieces, 0 .. MaxPieces] Of Tpiece; // pieces matrix

    // For optimization purpose pieces are of different sizes
    Nx, Ny: Integer; // number of pieces (width and height)
    K1, // zoom factor for pieces   1..6
    K2, // rectangle piece size (wihout outgrowth) 15*k1
    K3, // outgrowth size  4*k1

    Centrex, Centrey: Integer; // center of jigsaw image
    Origx, Origy: Integer; // top left image location
    Zrect: Trect; // rectangle image
    Ctrprior: Integer; // drawing priority

    // ini file fields
    Oldpuzzle: Boolean;
    OldPbwidth, OldPbheight: Integer;
    Oldorigx, Oldorigy: Integer;
    Oldnbp: Integer;
    Oldk1: Integer;
    Oldnbx: Integer;
    Oldnby: Integer;
    Oldzrect: Trect;
    Procedure Readinifile;
    Procedure Writeinifile;
    Function GetSpriteNumber(Spritenumber: Integer): String;
  Public
    Bmp0: Tbitmap; // read bitmap
    Nbp: Integer; // number of pieces = nx * ny used by FormGameOptions
    Firsttime: Boolean; // used by FormOpenPicture
    // Spritez unit
    Album: Talbum;
    Pool: Tpool;

    Procedure Log(LogRow: String);
    Procedure ToonLog;
    Procedure Menugris(Value: Boolean);
    Procedure Dimensions;
    Procedure Formes;
    Procedure PieceToRowCol;
    Procedure LoadfromOpenPicture;
    Procedure PuzzleDessign(Viewing: Boolean);
    Procedure Bez(Cancan: Tcanvas; Ox, Oy: Integer; AA: Array Of Tpoint);
    Procedure MakePuzzle(Nochange: Boolean);
    Procedure Fig(Cancan: Tcanvas; ARow, Acol: Integer);
    Function EnveloPartsArraye(ARow, Acol: Integer): Trect;
    Function Testfinal: Boolean;
    Procedure Autodeplace;
    Procedure PreviewViewing(Reduse: Boolean);
    Function CalcTime: String;
    Procedure ShowTime(DisplayTime: Boolean);
  End;

Var
  FormMain: TFormMain;

Implementation

{$R *.DFM}

Uses UIniOptions, UPreview, USettings, UAbout, UOptions;

Const
  // Bezier PieceCurves
  Courbe1: Array [0 .. BezierLineSteps] Of Tpoint = ((X: 00; Y: 00), (X: 01; Y: - 2), (X: 06; Y: 02), (X: 06; Y: 00), (X: 04;
    Y: - 6), (X: 11; Y: - 4), (X: 09; Y: 00), (X: 09; Y: 01), (X: 15; Y: - 1), (X: 15; Y: 00));
  Courbe2: Array [0 .. BezierLineSteps] Of Tpoint = ((X: 00; Y: 00), (X: 00; Y: 02), (X: 00; Y: 04), (X: 00; Y: 06), (X: 06;
    Y: 04), (X: 05; Y: 11), (X: 00; Y: 09), (X: 00; Y: 12), (X: 00; Y: 13), (X: 00; Y: 15));
  Courbe3: Array [0 .. BezierLineSteps] Of Tpoint = ((X: 00; Y: 00), (X: 00; Y: 01), (X: 06; Y: - 2), (X: 06; Y: 00), (X: 03;
    Y: 06), (X: 10; Y: 05), (X: 09; Y: 00), (X: 09; Y: - 2), (X: 15; Y: 00), (X: 15; Y: 00));
  Courbe4: Array [0 .. BezierLineSteps] Of Tpoint = ((X: 00; Y: 00), (X: 00; Y: 02), (X: 00; Y: 03), (X: 00; Y: 06), (X: - 6;
    Y: 04), (X: - 5; Y: 11), (X: 00; Y: 09), (X: 00; Y: 10), (X: 00; Y: 13), (X: 00; Y: 15));

Procedure TFormMain.FormCreate(Sender: TObject);

  Procedure MakeAlbum;
  Var
    BmpTemp: Tbitmap;
  Begin
    BmpTemp := Tbitmap.Create;
    Try
      BmpTemp.Width := Paintbox.Width;
      BmpTemp.Height := Paintbox.Height;
      BmpTemp.Canvas.Brush.Color := ClGradientInActiveCaption; // lightblue
      BmpTemp.Canvas.Fillrect(Rect(0, 0, BmpTemp.Width, BmpTemp.Height));
      BmpTemp.Canvas.Brush.Color := Clblack;
      Album := Talbum.Create(Paintbox.Canvas, BmpTemp); // create album
      PaintboxFlag := PfImage;
    Finally
      BmpTemp.Free;
    End;
  End;

Begin
//  default  paintbox   500 x 600
//           memowidth  300
//           mainform   598 x 919

  BoundsRect := IniOptions.Place;
  MemoLog.Clear;
  ToonLog;
  ProgressBar.Parent := StatusBar; // place the progress bar into the status bar
  StatusBarResize(Nil);
  StatusBar.Panels.Items[1].Text := '';
  StatusBar.Panels.Items[2].Text := '';
  StatusBar.Panels.Items[3].Text := '';
  LabelSolved.Visible := False;
  MenuSetting;

  // initial
  Firsttime := True;
  Oldpuzzle := False;
  PaintboxFlag := PfEmpty; // empty
  ImageIsLoaded := False;
  SpritesCreated := False;
  Moving := False;
  MoveAuto := False;
  Redo := False;
  Completed := False;
  Speed := 1;
  Startdir := GetCurrentdir;
  Menugris(True);
  MakeAlbum;

  Bmp0 := Tbitmap.Create;
  Bmp0.Width := 8;
  Bmp0.Height := 8;
  Bmp0.Pixelformat := Pf24bit;
  Bmp0.Canvas.Brush.Color := ClCream;
  Bmp0.Canvas.Fillrect(Rect(0, 0, Width, Height));

  Bmp1 := Tbitmap.Create;
  Bmp1.Assign(Bmp0);
  Bmp1.Canvas.Brush.Color := ClMoneyGreen;
  Bmp1.Canvas.Fillrect(Rect(0, 0, Width, Height));

  Log('Form create');
End;

Procedure TFormMain.FormActivate(Sender: TObject);
Begin
  Log('Activate-----------------------------');
  // MenuNewImageOudClick(sender);
//!! temp off for debugging
End;

Procedure TFormMain.FormClose(Sender: TObject; Var Action: TCloseAction);
Begin
  Log('Program closed');
  If IniOptions.AutoSolve Then
    MenuSolutionClick(Sender);
  IniOptions.Place := BoundsRect;
  TimerTime.Enabled := False;
  IniOptions.SaveToFile(IniOptions.IniFilename);
  Writeinifile;
  MemoLog.Lines.SaveToFile('Log\Log' + DateToStr(Now) + '.txt');
  Action := Cafree;
End;

Procedure TFormMain.FormDestroy(Sender: TObject);
Begin
  Log('Form destroy');
  Album.Free;
  Bmp0.Free;
  Bmp1.Free;
End;

// --------------------------------------------------------------------
Procedure TFormMain.Readinifile;
Var
  Num: TNumArray;
  { --- }

  Procedure Decodepuzzle;
  Var
    DRow, Dcol, Part: Integer;
    Partsleft: Boolean;
  Begin
    Log('Decode');

    Num := IniOptions.LoadIdentFromIniFile(CsIniScreenIdent);
    If Num[1] + Num[2] <> 0 Then
      Oldpuzzle := True;
    OldPbwidth := Num[1];
    OldPbheight := Num[2];
    Oldorigx := Num[3];
    Oldorigy := Num[4];
    Oldnbp := Num[5];
    Oldk1 := Num[6];

    Num := IniOptions.LoadIdentFromIniFile(CsIniPiecesIdent);
    Oldnbx := Num[1];
    If Oldnbx < 1 Then
      Oldnbx := 1;
    If Oldnbx > MaxPieces Then
      Oldnbx := 1;
    Oldnby := Num[2];
    If Oldnby < 1 Then
      Oldnby := 1;
    If Oldnby > MaxPieces Then
      Oldnby := 1;
    Oldzrect := Rect(Num[3], Num[4], Num[5], Num[6]);

    Num := IniOptions.LoadIdentFromIniFile(CsIniTempIdent);
    Timebefore := Encodetime(Word(Num[1]), Word(Num[2]), Word(Num[3]), 0);

    // all pieces
    Partsleft := True;
    Part := 1;
    While Partsleft = True Do
      Begin
        Num := IniOptions.LoadIdentFromIniFile(CsIniPartIdent + Inttostr(Part));
        // Number6;
        Partsleft := (Num[1] + Num[2] <> 0); // no pieces left
        If Partsleft Then
          Begin
            DRow := Num[1] Div Oldnbx;
            Dcol := Num[1] Mod Oldnbx;
            PartsArray[DRow, Dcol].Pleft := Num[2] Div 1000;
            PartsArray[DRow, Dcol].Ptop := (Num[2] Mod 1000) Div 100;
            PartsArray[DRow, Dcol].Pright := (Num[2] Mod 100) Div 10;
            PartsArray[DRow, Dcol].Pbottom := Num[2] Mod 10;

            PartsArray[DRow, Dcol].Pinit.X := Num[3];
            PartsArray[DRow, Dcol].Pinit.Y := Num[4];

            PartsArray[DRow, Dcol].Pflag := Num[5];
            PartsArray[DRow, Dcol].PRotation := Num[6];
            PartsArray[DRow, Dcol].PRotation0 := Num[6];
          End;
        Inc(Part);
      End;
  End;

Begin
  Log('Read Inifile'); // earlier jigsaw
  If FileExists(IniOptions.Picture) Then // if jigsaw stored
    Begin
      Decodepuzzle; // get previous screen parameters and pieces coordinates
    End
  Else
    IniOptions.CleanupIniFile; // remove oldsettings (waste) from ini;
End;

Procedure TFormMain.MenuRotatePiecesClick(Sender: TObject);
Begin
  MenuRotatePieces.Checked := Not MenuRotatePieces.Checked;
  IniOptions.Rotatepieces := MenuRotatePieces.Checked;
End;

Procedure TFormMain.LoadfromOpenPicture;

Begin
  Log('Load from OpenPicture');
  If IniOptions.Picture = '' Then // no picture saved
    Begin
      If OpenPictureDialog.Execute Then
        IniOptions.Picture := OpenPictureDialog.FileName; // select picture
    End;
  If IniOptions.Picture <> '' Then // picture selected
    If FileExists(IniOptions.Picture) Then // picture exits
      Begin
        Log('# OpenPicture ' + IniOptions.Picture);
        FormPreview.LoadFromPicture; // store also in preview
        if not assigned(Bmp0) then
          Bmp0 := Tbitmap.Create;
        With FormPreview.ImagePreview.Picture do
          begin
            Bmp0.Width := Graphic.Width;
            Bmp0.Height := Graphic.Height;
            Bmp0.Canvas.Draw(0, 0, Graphic);
          end;
        FormPreview.CalcImageScaling;
        StatusBar.Panels.Items[2].Text := 'Size ' + Inttostr(Bmp0.Width) + ' x ' + Inttostr(Bmp0.Height);
      End;
End;

Procedure TFormMain.StatusBarResize(Sender: TObject);
Var
  R: Trect;
Begin
  // place the progress bar over the 3nd statusbarpanel
  SendMessage(StatusBar.Handle, SB_GETRECT, 2, LPARAM(@R));
  ProgressBar.SetBounds(R.Left, R.Top, R.Right, R.Bottom);
End;

Procedure TFormMain.Writeinifile;
Var
  S: String;
  Ctri, I, J: Integer;
  H, M, Sec, Ms: Word;
Begin
  Log('Write inifile');
  If ImageIsLoaded And (Completed = False) Then
    Begin
      IniOptions.SaveIdentToIniFile(CsIniScreenIdent, Inttostr(Paintbox.Width) + ';' + Inttostr(Paintbox.Height) + ';' +
        Inttostr(Origx) + ';' + Inttostr(Origy) + ';' + Inttostr(Nbp) + ';' + Inttostr(K1) + ';');
      IniOptions.SaveIdentToIniFile(CsIniPiecesIdent, Inttostr(Nx) + ';' + Inttostr(Ny) + ';' + Inttostr(Zrect.Left) + ';' +
        Inttostr(Zrect.Top) + ';' + Inttostr(Zrect.Right) + ';' + Inttostr(Zrect.Bottom) + ';');
      Decodetime(Now - StartTime, H, M, Sec, Ms);
      IniOptions.SaveIdentToIniFile(CsIniTempIdent, Inttostr(H) + ';' + Inttostr(M) + ';' + Inttostr(Sec) + ';0;0;0');
      // pieces
      Ctri := 0;
      For J := 0 To Ny - 1 Do
        For I := 0 To Nx - 1 Do
          Begin
            Inc(Ctri);
            S := Inttostr(Ctri) + ';';
            With PartsArray[J, I] Do
              Begin
                S := S + Inttostr(Pleft) + Inttostr(Ptop) + Inttostr(Pright) + Inttostr(Pbottom) + ';';
              End;
            S := S + Inttostr(Pool.Getposx(Ctri)) + ';'; // position x
            S := S + Inttostr(Pool.Getposy(Ctri)) + ';'; // position y
            S := S + Inttostr(Pool.Gettag(Ctri)) + ';'; // tag value
            PieceNr := Ctri;
            PieceToRowCol;
            S := S + Inttostr(PartsArray[PieceRow, PieceCol].PRotation) + ';'; // rotation
            IniOptions.SaveIdentToIniFile(CsIniPartIdent + Inttostr(Ctri), S); // store part in ini
          End;
    End;
End;

Procedure TFormMain.MenuExitClick(Sender: TObject);
Begin
  Close;
End;

Procedure TFormMain.Menugris(Value: Boolean);
Begin
  If Value Then
    Log('memugris aan')
  Else
    Log('memugris uit');
  MenuShowImage.Enabled := Not Value;
  MenuSolution.Enabled := Not Value;
End;

Procedure TFormMain.MenuNewStartClick(Sender: TObject);
Var
  N, Row, Col: Integer;
Begin
  Log('menu new start click');
  IF PaintboxFlag = PfImage Then // bmp1
    Exit;
  CountAutoMove := 0;
  CountMouseMove := 0;
  Completed := False;
  ShowTime(False);
  MenuShowImage.Caption := '&Display image';
  Paintbox.Invalidate;
  N := 0;
  For Row := 0 To Ny - 1 Do // do for all parts
    For Col := 0 To Nx - 1 Do
      Begin
        Inc(N);
        Pool.Settoon(N, 'P' + Inttostr(N));
        Pool.Setpos(N, PartsArray[Row, Col].Pinit.X, PartsArray[Row, Col].Pinit.Y);
        Pool.Setpriority(N, -N);
        Pool.Setanimstyle(N, PartsArray[Row, Col].PRotation0);
        Pool.Settag(N, 0);
        PartsArray[Row, Col].PRotation := PartsArray[Row, Col].PRotation0;
      End;
  StatusBar.Panels.Items[1].Text := ' ' + Inttostr(Nbp) + ' pieces';
  Pool.Animate; // paint screen
  StartTime := Now;
  Menugris(False);
  ShowTime(True);
End;

Procedure TFormMain.MenuNewImageClick(Sender: TObject);
// Read a stored image and paraneters, a get a new image
Var
  BmpBackGround: Tbitmap;
  Choice: Integer;

  Function CompareDimensions: Boolean;
  Begin
    Result := (OldPbwidth = Paintbox.Width) AND (OldPbheight = Paintbox.Height);
    Result := Result AND (Oldorigx = Origx) AND (Oldorigy = Origy);
    Result := Result AND (Oldnbp = Nbp) AND (Oldk1 = K1) AND (Oldnbx = Nx) AND (Oldnby = Ny);
    Result := Result AND (Oldzrect.Left = Zrect.Left) AND (Oldzrect.Top = Zrect.Top);
    Result := Result AND (Oldzrect.Right = Zrect.Right) AND (Oldzrect.Bottom = Zrect.Bottom);
  End;

Begin
  Log('menu new image click');
  CountAutoMove := 0;
  CountMouseMove := 0;
  Completed := False;
  ShowTime(False);
  MenuShowImage.Caption := '&Display image';
  If Firsttime Then
    Begin
      Readinifile; // earlier jigsaw
      If Oldpuzzle AND (Messagedlg('Continue the game "' + ExtractFileName(IniOptions.Picture) +
        '" not achieved last time ?', MtConfirmation, [MbYes, MbNo], 0) = MrYes) Then
        Begin
          LoadfromOpenPicture;
          ImageIsLoaded := True;
          Dimensions;
          If CompareDimensions Then // compare measurements
            Begin
              Redo := True;
              MakePuzzle(True);
              StartTime := StartTime - Timebefore;
              Redo := False;
              ShowTime(IniOptions.ShowTime);
              PaintboxFlag := PfPart;
              Firsttime := False;
              Oldpuzzle := False;
              Exit;
            End
          Else
            Begin
              Showmessage('Display device or Image settings are not the same.' + ' Impossible to restart previous game.');
            End;
        End;
      Firsttime := False;
      ImageIsLoaded := True;
    End;
  SpritesCreated := False;
  BmpBackGround := Album.Getareabmp; // pointer to the background bitmap
  With BmpBackGround.Canvas Do
    Begin
      Brush.Color := ClGradientInActiveCaption; // backgroundcolor
      Fillrect(Rect(0, 0, Paintbox.Width, Paintbox.Height));
    End;

  Repeat
    If OpenPictureDialog.Execute Then
      Begin
        IniOptions.Picture := OpenPictureDialog.FileName;
        LoadfromOpenPicture;
        FormPreview.ImagePreview.Picture.LoadFromFile(IniOptions.Picture);
        Bmp1.Width := FormPreview.ImagePreview.Picture.Graphic.Width;
        Bmp1.Height := FormPreview.ImagePreview.Picture.Graphic.Height;
        Bmp1.Canvas.Draw(0, 0, FormPreview.ImagePreview.Picture.Graphic);
        ImageIsLoaded := True;
        PaintboxFlag := PfImage; // bmp1
      End;

    Dimensions;
    Formes;
    PuzzleDessign(True);
    Paintbox.Canvas.Draw(Origx, Origy, Bmp1);
    Log('Option showmodal');
    Choice := FormGameOptions.Showmodal;

  Until (Choice = Mrok) Or (Choice = Mrcancel);

  IF Choice = Mrok Then
    Begin
      Menugris(False);
      PuzzleDessign(False);
      MakePuzzle(False);
      MenuStartAgain.Enabled := True;
      ShowTime(IniOptions.ShowTime);
      PaintboxFlag := PfPart; // parts
    End
  Else
    Begin
      PaintboxFlag := PfImage; // bmp1
      Paintbox.Invalidate;
      ImageIsLoaded := False;
    End;
End;

Procedure TFormMain.MenuNiveauClick(Sender: TObject);
Begin
  // Image size
  Log('& radio group niveau');
  If (Sender Is TMenuItem) Then
    With Sender As TMenuItem Do
      IniOptions.Niveau := Tag;
  MenuSetting;

  Dimensions;
  Formes;
  PuzzleDessign(True);
End;

Procedure TFormMain.Dimensions;
Var
  W, H: Integer; // bitmap image puzzle
  W1, H1: Single; // original image size
  W2, H2: Single; // paintbox size - margin
  K, Kw, Kh: Single;
Begin
  Log('# Dimensions');
  IF Not ImageIsLoaded Then
    Exit;
  // Piece unit size is 15x15. A borderpiece is 11*11 (K2-K3)
  K1 := IniOptions.Niveau + 3;
//  if BmpBackGround.Width > Paintbox.Width then
//    K1 := K1 - 1;
  K2 := K1 * 15;
  K3 := K1 * 4;
  W1 := Bmp0.Width;    // image
  H1 := Bmp0.Height;   // image
  W2 := Paintbox.Width;
  If H1 >= W1 Then
    H2 := Paintbox.Height - 24 - K2 - K3 - K3
  Else
    H2 := Paintbox.Height - ((K2 + K3) * 3);
  Paintbox.Canvas.FloodFill(10, 10, ClYellow, FsSurface);
  Kw := 1;
  Kh := 1;
  If W1 > W2 Then      // imaqe.width > paintbox.width
    Kw := W2 / W1;
  If H1 > H2 Then      // imaqe.heigth > paintbox.heigth
    Kh := H2 / H1;
  If Kw < Kh Then      // width % <  heigth %
    K := Kw
  Else
    K := Kh;
  W := Round(W1 * K);   // K = scalingfactor
  H := Round(H1 * K);
  Nx := W Div K2; // number of x pieces
  Ny := H Div K2; // number of y pieces
  If Nx < 1 Then
    Nx := 1;
  If Ny < 1 Then
    Ny := 1;
  Inc(Nx, 2); // +2 for end pieces
  Inc(Ny, 2); // +2
  Nbp := Nx * Ny; // total pieces number
  Ctrprior := Nbp;
  With Zrect Do // jigsaw Rectangle
    Begin
      Left := 0;
      Top := 0;
      Right := Nx * K2 - K3 * 2;
      Bottom := Ny * K2 - K3 * 2;
    End;
  Bmp1.Free;
  Bmp1 := Tbitmap.Create;
  Bmp1.Width := Zrect.Right - Zrect.Left;
  Bmp1.Height := Zrect.Bottom - Zrect.Top;
  Centrex := Paintbox.Width Div 2; // center
  Centrey := Paintbox.Height Div 2;
  Origx := Centrex - Bmp1.Width Div 2; // left,top
  Origy := Centrey - Bmp1.Height Div 2;
  Bmp1.Canvas.Stretchdraw(Zrect, Bmp0); // bmp1 = image to be cut into pieces, draws Bmp1 to image1
  StatusBar.Panels.Items[1].Text := ' ' + Inttostr(Nbp) + ' pieces';
  PreviewViewing(False);
End;

Procedure TFormMain.MenuAboutExecute(Sender: TObject);
Begin
  AboutBox.Showmodal;
End;

Procedure TFormMain.MenuHelpExecute(Sender: TObject);
Begin
  Application.HelpContext(IDH_HELP);
End;

Procedure TFormMain.PreviewViewing(Reduse: Boolean);
Begin
  Log('# Preview Viewing on/off');
  FormPreview.Visible := Not FormPreview.Visible;
End;

// random pieces shapes
Procedure TFormMain.Formes;
Var
  Col, Row: Integer;
  Z: Word;

  Procedure ClearArray;
  Var
    Col, Row: Integer;
  Begin
    For Row := 0 To MaxPieces Do
      For Col := 0 To MaxPieces Do
        Begin
          PartsArray[Row, Col].Pleft := 0;
          PartsArray[Row, Col].Ptop := 0;
          PartsArray[Row, Col].Pright := 0;
          PartsArray[Row, Col].Pbottom := 0;

          PartsArray[Row, Col].Pinit.X := 0;
          PartsArray[Row, Col].Pinit.Y := 0;
          PartsArray[Row, Col].PRotation := 0;
          PartsArray[Row, Col].PRotation0 := 0;
          PartsArray[Row, Col].Pflag := 0;
        End;
  End;

Begin
  Log('# Formes');
  ClearArray;
  For Row := 0 To Ny - 1 Do
    Begin
      For Col := 0 To Nx - 1 Do
        Begin
          // horizontal
          If Row = 0 Then
            Begin
              PartsArray[Row, Col].Ptop := 0; // top left border
            End
          Else
            Begin
              If Row = Ny Then
                Begin
                  PartsArray[Row - 1, Col].Pbottom := 0 // bottom left border
                End
              Else
                Begin
                  Z := Random(2);
                  If Z = 0 Then
                    Begin
                      PartsArray[Row, Col].Ptop := 1;
                      PartsArray[Row - 1, Col].Pbottom := 1;
                    End
                  Else
                    Begin
                      PartsArray[Row, Col].Ptop := 2;
                      PartsArray[Row - 1, Col].Pbottom := 2;
                    End;
                End;
            End;
          // Vertical
          IF Col = 0 Then
            Begin
              PartsArray[Row, Col].Pleft := 0; // top right corner
            End
          Else
            Begin
              If Col = Nx Then
                Begin
                  PartsArray[Row, Col - 1].Pright := 0; // bottom right corner
                End
              Else
                Begin
                  Z := Random(2);
                  If Z = 0 Then
                    Begin
                      PartsArray[Row, Col].Pleft := 1;
                      PartsArray[Row, Col - 1].Pright := 1;
                    End
                  Else
                    Begin
                      PartsArray[Row, Col].Pleft := 2;
                      PartsArray[Row, Col - 1].Pright := 2;
                    End;
                End;
            End;
        End; // For col
    End; // For Row
End;

procedure TFormMain.FormResize(Sender: TObject);
begin
  if assigned(FormPreview) then
    if FormPreview.Showing then // and onscreen
      FormPreview.UpdatePosition;
end;

Var
  StoredRow: String;

Procedure TFormMain.Log(LogRow: String);
Begin
  If LogRow <> StoredRow Then // same as lastone
    begin
      MemoLog.Lines.Add(DateTimeToStr(Now) + ' : ' + LogRow);
      StoredRow := LogRow;
    end;
End;

Procedure TFormMain.PieceToRowCol;
Begin
  PieceRow := (PieceNr - 1) Div Nx;
  PieceCol := (PieceNr - 1) Mod Nx;
End;

Procedure Ligne(Cancan: Tcanvas; X1, Y1, X2, Y2: Integer);
Begin
  FormMain.Log('# Ligne');
  With Cancan Do
    Begin
      Moveto(X1, Y1);
      Lineto(X2, Y2);
    End;
End;

// polybezier API windows (uses the canvas handle)
Procedure TFormMain.Bez(Cancan: Tcanvas; Ox, Oy: Integer; AA: Array Of Tpoint);
Var
  N: Integer;
  HH: Thandle;
  BB: Array [0 .. BezierLineSteps] Of Tpoint; // drawing pieces

Begin
  For N := 0 To BezierLineSteps Do
    Begin
      BB[N].X := AA[N].X * K1 + Ox;
      BB[N].Y := AA[N].Y * K1 + Oy;
    End;
  HH := Cancan.Handle;
  Winapi.Windows.Polybezier(HH, BB, 10);
End;

Procedure TFormMain.MenuSetting;
Begin
  MenuRotatePieces.Checked := IniOptions.Rotatepieces;
  MenuNiveau4.Checked := (MenuNiveau4.Tag = IniOptions.Niveau);
  MenuNiveau3.Checked := (MenuNiveau3.Tag = IniOptions.Niveau);
  MenuNiveau2.Checked := (MenuNiveau2.Tag = IniOptions.Niveau);
  MenuNiveau1.Checked := (MenuNiveau1.Tag = IniOptions.Niveau);
End;

Procedure TFormMain.MenuSettingsClick(Sender: TObject);
Begin
  FormSettings.Showmodal;
  ToonLog;
End;

Procedure TFormMain.MenuBorderColorClick(Sender: TObject);
Begin
  Log('FormMain Bordercolor');
  With FormGameOptions do // show only the bordercolor
    Begin
      ColorBox.Selected := IniOptions.PieceBorderColor; // current
      RadioGroupNiveau.Visible := False;
      CheckBoxRotate.Visible := False;
      PnlRight.Align := Albottom;
      PnlRight.AutoSize := True;
      GroupBoxBorderColor.Align := alTop;
      FormGameOptions.AutoSize := True;
      if Showmodal = Mrok then
        begin
          IniOptions.PieceBorderColor := ColorBox.Color;
          PuzzleDessign(True); // bordercolor changed
        End;

    End;
End;

Procedure TFormMain.PuzzleDessign(Viewing: Boolean);
Var
  Col, Row: Integer;
Begin
  Log('FormMain PuzzleDesign');
  Album.Drawarea(Rect(0, 0, Paintbox.Width, Paintbox.Height));
  With Bmp1.Canvas Do
    Begin
      Stretchdraw(Zrect, Bmp0);
      Pen.Color := IniOptions.PieceBorderColor;
      Pen.Width := 2; // borderwidth  test
      Moveto(0, 0);
      Lineto(Bmp1.Width - 1, 0);
      Lineto(Bmp1.Width - 1, Bmp1.Height - 1);
      Lineto(0, Bmp1.Height - 1);
      Lineto(0, 0);
    End;

  Log('FormMain create partsknots');
  For Row := 0 To Ny - 1 Do
    Begin
      For Col := 0 To Nx - 1 Do
        Begin
          Case PartsArray[Row, Col].Ptop Of
            1:
              Bez(Bmp1.Canvas, Col * K2 - K3, Row * K2 - K3, Courbe1);
            2:
              Bez(Bmp1.Canvas, Col * K2 - K3, Row * K2 - K3, Courbe3);
          End;
          Case PartsArray[Row, Col].Pleft Of
            1:
              Bez(Bmp1.Canvas, Col * K2 - K3, Row * K2 - K3, Courbe2);
            2:
              Bez(Bmp1.Canvas, Col * K2 - K3, Row * K2 - K3, Courbe4);
          End;
        End; // for col
    End; // for Row
  If Viewing Then
    Paintbox.Canvas.Draw(Origx, Origy + 1, Bmp1);
End;

// Create pieces toons a associated sprites
Procedure TFormMain.MakePuzzle(Nochange: Boolean);
Const
  // Black pixels recplaced with dark pixels (Mask is black)
  JigsawBlack = $00353535; // darkgray

Var
  SaveCursor: TCursor;
  Col, Row: Integer;
  Nbimg: Integer;
  R1, R2: Trect;
  P: Tpoint;
  BmpX: Tbitmap;
  Csx, Csy: Integer;
  Nomtoon: String;
  BmpBackGround: Tbitmap;
  // Xmask: Tcolor;
  PartRotation: Integer; // pieces rotation
  Bestmarge: Integer;

  Function GetRandomPositions: Tpoint;
  Var
    I: Integer;
    P: Tpoint;
  Begin
    I := Rndi(0, 9);
    P.X := Rndi(0, Paintbox.Width - K2 - K3);
    P.Y := Rndi(0, Paintbox.Height - K2 - K3);
    If Ptinrect(R1, P) = True Then
      Begin
        If P.X < Centrex - K2 Then
          Begin
            Case Bestmarge Of
              0:
                If I Mod 2 = 0 Then
                  P.X := Rndi(0, R1.Left);
              1:
                P.X := Rndi(0, R1.Left);
              3:
                If I Mod 2 = 0 Then
                  P.X := -K2;
            End;
          End
        Else
          Begin
            Case Bestmarge Of
              0:
                If I Mod 2 = 0 Then
                  P.X := Rndi(R1.Right, Paintbox.Width - K2 - K3);
              1:
                P.X := Rndi(R1.Right, Paintbox.Width - K2 - K3);
              3:
                If I Mod 2 = 0 Then
                  P.X := Paintbox.Width - K2 - K3;
            End;
          End;
        If P.Y < Centrey - K2 Then
          Begin
            Case Bestmarge Of
              0:
                If I Mod 2 = 1 Then
                  P.Y := Rndi(0, R1.Top - K3);
              2:
                P.Y := Rndi(0, R1.Top - K3);
              3:
                If I Mod 2 = 1 Then
                  P.Y := -K2;
            End;
          End
        Else
          Begin
            Case Bestmarge Of
              0:
                If I Mod 2 = 1 Then
                  P.Y := Rndi(R1.Bottom, Paintbox.Height - K2 - K3);
              2:
                P.Y := Rndi(R1.Bottom, Paintbox.Height - K2 - K3);
              3:
                If I Mod 2 = 1 Then
                  P.Y := Paintbox.Height - K2;
            End;
          End;
      End;
    Result.X := (P.X Div 16) * 16 + 8;
    Result.Y := (P.Y Div 16) * 16 + 8;
  End;

Begin
  Log('FormMain Make puzzle');
  SaveCursor := Screen.Cursor;
  Try
    Screen.Cursor := Crhourglass;

    BmpBackGround := Album.Getareabmp; // pointer to album background bitmap
    With BmpBackGround.Canvas Do
      Begin
        Pen.Color := Clred;
        Brush.Color := $00CBCB96; // cyan
        Rectangle(Origx - 2, Origy - 1, Origx + Bmp1.Width + 2, Origy + Bmp1.Height + 4);
      End;

    Album.Drawarea(Rect(0, 0, Paintbox.Width, Paintbox.Height)); // erase

    Bmp0.Canvas.Brush.Color := JigsawBlack;
    Bmp1.Canvas.Brush.Color := JigsawBlack;
    Bmp1.Canvas.Brushcopy(Zrect, Bmp0, Rect(0, 0, Bmp0.Width, Bmp0.Height), Clblack);
    Bmp1.Canvas.Brush.Color := Clblack;
    Bmp1.Pixelformat := Pf24bit;

    ProgressBar.Position := 0;
    ProgressBar.Max := Nx * Ny * 2;
    Nbimg := 1;
    Csx := Bmp1.Width Div 2;
    Csy := Bmp1.Height Div 2;
    For Row := 0 To Ny - 1 Do
      Begin
        For Col := 0 To Nx - 1 Do
          Begin
            R1 := EnveloPartsArraye(Row, Col);
            BmpX := Tbitmap.Create;
            Try
              BmpX.Width := R1.Right - R1.Left + 1;
              BmpX.Height := R1.Bottom - R1.Top + 1;
              BmpX.Pixelformat := Pf24bit;
              R2 := Rect(0, 0, BmpX.Width, BmpX.Height);
              With BmpX.Canvas Do
                Begin
                  // create a white mask
                  Copymode := CmSrcCopy; // normal
                  Brush.Color := Clblack;
                  Fillrect(R2); // fill part
                  Pen.Color := ClWhite;
                  Fig(BmpX.Canvas, Row, Col);

                  Brush.Color := ClWhite;
                  FloodFill(BmpX.Width Div 2 + 2, BmpX.Height Div 2 + 2, ClWhite, Fsborder);
                  Copymode := CmSrcAnd;
                  Copyrect(R2, Bmp1.Canvas, R1);
                  Pen.Color := IniOptions.PieceBorderColor;
                  Fig(BmpX.Canvas, Row, Col);
                End;
              Nomtoon := GetSpriteNumber(Nbimg);
              Album.CreateToon(Nomtoon, BmpX, 1, 1, 1, BmpX.Width, BmpX.Height, 0, Clblack);
              Album.Settooncenter(Nomtoon, R1.Left - Csx, R1.Top - Csy);
              Inc(Nbimg);
            Finally
              BmpX.Free;
            End;
          End; // for col
      End; // for Row

    // pool and sprites
    Pool := Album.Createpool(Nbimg);

    // rectangle to exclude pieces
    R1 := Rect(Origx - K2 - K3 * 2, Origy - K2 - K3 * 2, Origx + Bmp1.Width, Origy + Bmp1.Height);
    Bestmarge := 0;
    If Origy < K2 + K3 Then
      Bestmarge := 1;
    If Origx < K2 + K3 Then
      Bestmarge := 2;
    If (Origx < K2 + K3) And (Origy < K2 + K3) Then
      Bestmarge := 3;
    Nbimg := 0;
    For Row := 0 To Ny - 1 Do
      Begin
        For Col := 0 To Nx - 1 Do
          Begin
            Inc(Nbimg);
            ProgressBar.Position := Nx * Ny + Nbimg;
            // Nomtoon := GetSpriteNumber(Nx * Ny + Nbimg);
            Nomtoon := GetSpriteNumber(Nbimg);
            Pool.Createsprite(Nbimg, Nomtoon);
            If Redo Then // previous locations
              Begin
                P.X := PartsArray[Row, Col].Pinit.X;
                P.Y := PartsArray[Row, Col].Pinit.Y;
                Pool.Settag(Nbimg, PartsArray[Row, Col].Pflag);
              End
            Else
              Begin // new puzzle, random locations
                P := GetRandomPositions;
              End;
            Pool.Setpos(Nbimg, P.X, P.Y);
            Pool.Setpriority(Nbimg, -Nbimg);
            PartsArray[Row, Col].Pinit.X := P.X;
            PartsArray[Row, Col].Pinit.Y := P.Y;

            If Nochange Then
              PartRotation := PartsArray[Row, Col].PRotation0
            Else
              If IniOptions.Rotatepieces = True Then // with rotation
                PartRotation := Rndi(0, 3) // random rotatation
              Else
                PartRotation := 0;
            Pool.Setanimstyle(Nbimg, PartRotation);
            PartsArray[Row, Col].PRotation := PartRotation;
            PartsArray[Row, Col].PRotation0 := PartRotation;
          End;
      End;
    Pool.Animate; // show pieces using drawing priorities
    SpritesCreated := True;
    Menugris(False);
    StartTime := Now;
  Finally
    Screen.Cursor := SaveCursor;
  End;
  ProgressBar.Position := 0;
End;

// draws a piece on BmPartsArray
Procedure TFormMain.Fig(Cancan: Tcanvas; ARow, Acol: Integer);
Var
  Xx, Yy: Integer;
Begin
  Log('Fig');
  Xx := 0;
  Yy := 0;
  Case PartsArray[ARow, Acol].Ptop Of
    0:
      Yy := K3;
    1:
      Yy := -K3 - 2;
    2:
      Yy := -K1 - 1;
  End;
  Case PartsArray[ARow, Acol].Pleft Of
    0:
      Xx := K3;
    1:
      Xx := -2;
    2:
      Xx := -K3 - 2;
  End;
  Yy := -Yy;
  Xx := -Xx;
  Case PartsArray[ARow, Acol].Ptop Of
    0:
      Ligne(Cancan, Xx, 0, Xx + K2, 0);
    1:
      Bez(Cancan, Xx, Yy, Courbe1);
    2:
      Bez(Cancan, Xx, Yy, Courbe3);
  End;
  Case PartsArray[ARow, Acol].Pbottom Of
    0:
      Ligne(Cancan, Xx, Yy + K2 - K3, Xx + K2, Yy + K2 - K3);
    1:
      Bez(Cancan, Xx, Yy + K2, Courbe1);
    2:
      Bez(Cancan, Xx, Yy + K2, Courbe3);
  End;
  Case PartsArray[ARow, Acol].Pleft Of
    0:
      Ligne(Cancan, 0, Yy, 0, Yy + K2);
    1:
      Bez(Cancan, Xx, Yy, Courbe2);
    2:
      Bez(Cancan, Xx, Yy, Courbe4);
  End;
  Case PartsArray[ARow, Acol].Pright Of
    0:
      Ligne(Cancan, Xx + K2 - K3, Yy, Xx + K2 - K3, Yy + K2);
    1:
      Bez(Cancan, Xx + K2, Yy, Courbe2);
    2:
      Bez(Cancan, Xx + K2, Yy, Courbe4);
  End;
End;

Function TFormMain.EnveloPartsArraye(ARow, Acol: Integer): Trect;
Var
  Dx1, Dy1, Dx2, Dy2: Integer;
Begin
  Log('EnveloPartsArraye');
  Dx1 := 0;
  Dx2 := 0;
  Dy1 := 0;
  Dy2 := 0;
  Case PartsArray[ARow, Acol].Ptop Of
    0:
      Dy1 := K3;
    1:
      Dy1 := -K3 - 2;
    2:
      Dy1 := -K1 - 1;
  End;
  Case PartsArray[ARow, Acol].Pbottom Of
    0:
      Dy2 := K2 - K3;
    1:
      Dy2 := K2 + K1 + 2;
    2:
      Dy2 := K2 + K3 + 2;
  End;
  Case PartsArray[ARow, Acol].Pleft Of
    0:
      Dx1 := K3;
    1:
      Dx1 := -2;
    2:
      Dx1 := -K3 - 2;
  End;
  Case PartsArray[ARow, Acol].Pright Of
    0:
      Dx2 := K2 - K3;
    1:
      Dx2 := K2 + K3 + 2;
    2:
      Dx2 := K2 + 2;
  End;
  Result.Left := Acol * K2 - K3 + Dx1;
  Result.Top := ARow * K2 - K3 + Dy1;
  Result.Right := Acol * K2 - K3 + Dx2;
  Result.Bottom := ARow * K2 - K3 + Dy2;
End;

Procedure TFormMain.PaintboxPaint(Sender: TObject);
Begin
  Log('Paintboxpaint');
  If PaintboxFlag = PfEmpty Then // empty
    Exit;
  Album.Drawarea(Rect(0, 0, Paintbox.Width, Paintbox.Height));
  If PaintboxFlag = PfImage Then // bmp1
    Paintbox.Canvas.Draw(Origx, Origy, Bmp1);
  If SpritesCreated Then
    Pool.Animate;
End;

Procedure TFormMain.MenuShowImageClick(Sender: TObject);
Begin
  FormPreview.Visible := Not FormPreview.Visible;

  If MenuShowImage.Caption = '&Hide image' Then
    Begin
      MenuShowImage.Caption := '&Show image';
    End
  Else
    Begin
      IF IniOptions.Size = 0 Then
        PreviewViewing(False)
      Else
        PreviewViewing(True);
      MenuShowImage.Caption := '&Hide image';
    End;
End;

Procedure TFormMain.PaintboxMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
Var
  P: Tpoint;
Begin
  Log('Paintbox mouse down');

  IF MoveAuto Then
    Exit;
  IF IniOptions.AutoSolve Then
    Exit;
  P.X := X;
  P.Y := Y;
  PieceNr := 0;
  If Pool = Nil Then
    Exit;
  With Pool Do
    Begin
      PieceNr := Ptinsprite(P, -4);
      If PieceNr > 0 Then
        Begin
          Inc(Ctrprior);
          SetpriorityLast(PieceNr, Ctrprior);
          // If (Button = Mbright) And (FormSettings.RadioGroupRotate.Itemindex = 1) Then
          If (Button = Mbright) And (IniOptions.Rotatepieces = True) Then // right button = rotation
            Begin
              PieceToRowCol;
              Inc(PartsArray[PieceRow, PieceCol].PRotation);
              If PartsArray[PieceRow, PieceCol].PRotation > 3 Then
                PartsArray[PieceRow, PieceCol].PRotation := 0;
              Setanimstyle(PieceNr, PartsArray[PieceRow, PieceCol].PRotation);
              Animatemono(PieceNr);
            End
          Else
            Begin // automatic move with <Shift> or <Ctrl>
              If (Ssctrl In Shift) Or (Ssshift In Shift) Then
                Begin
                  Autodeplace;
                  Testfinal;
                End
              Else
                Begin
                  Moving := True;
                  Mx := X - Getposx(PieceNr);
                  My := Y - Getposy(PieceNr);
                End;
            End;
        End
      Else
        PieceNr := 0;
    End;
End;

Procedure TFormMain.PaintboxMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
Begin
  If MoveAuto Then
    Exit;
  If Not Moving Then
    Exit;
  Log('Paintbox mouse move');
  With Pool Do
    Begin
      // replace piece
      If X - Mx + Getdcx(PieceNr) < 0 Then
        X := -Getdcx(PieceNr) + Mx;
      IF Y - My + Getdcy(PieceNr) < 0 Then
        Y := -Getdcx(PieceNr) + My;
      If X - Mx + Getdcx(PieceNr) > Paintbox.Width Then
        X := Paintbox.Width - Getdcx(PieceNr) + Mx;
      If Y - My + Getdcy(PieceNr) > Paintbox.Height Then
        Y := Paintbox.Height - Getdcy(PieceNr) + My;
      Setpos(PieceNr, X - Mx, Y - My);
      // show piece
      Animatemono(PieceNr);
    End;
End;

Procedure TFormMain.PaintboxMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
Begin
  Log('Paintbox mouse up');
  If MoveAuto Then
    Exit;
  If Not Moving Then
    Exit;
  Moving := False;
  PieceToRowCol;
  With Pool Do
    Begin
      If (Abs(Getposx(PieceNr) - Centrex - Getoffsetx(PieceNr)) < IniOptions.Magnet) AND
        (Abs(Getposy(PieceNr) - Centrey - Getoffsety(PieceNr)) < IniOptions.Magnet) AND
        (PartsArray[PieceRow, PieceCol].PRotation = 0) Then
        Begin
          Setpos(PieceNr, Centrex + Getoffsetx(PieceNr), Centrey + Getoffsety(PieceNr));
          Setpriority(PieceNr, 0); // low priority
          Settag(PieceNr, 1); // final location flag
          Animatemono(PieceNr);
        End
      Else
        Begin // moved from final location
          SetpriorityLast(PieceNr, Ctrprior);
          Settag(PieceNr, 0); // remove final location flag
        End;
      PieceNr := 0;
    End;
  Inc(CountMouseMove);
  Testfinal;
End;

Function TFormMain.CalcTime: String;
Begin
  Result := ' ' + TimeToStr(Now - StartTime);
End;

Procedure TFormMain.TimerTimeTimer(Sender: TObject);
Begin
  // Show time left at Statusbar
  StatusBar.Panels.Items[0].Text := CalcTime;
End;

procedure TFormMain.ToonLog;
begin
  // show log
  MemoLog.Visible := IniOptions.ShowLog;
  if MemoLog.Visible then
    FormMain.Width := FormMain.Width + MemoLog.Width
  else
    FormMain.Width := FormMain.Width - MemoLog.Width;
end;

Procedure TFormMain.ShowTime(DisplayTime: Boolean);
Begin
  TimerTime.Enabled := True;
  If Not DisplayTime Then
    StatusBar.Panels.Items[0].Text := '';
End;

Function TFormMain.GetSpriteNumber(Spritenumber: Integer): String;
Begin
  ProgressBar.Position := Spritenumber;
  StatusBar.Panels.Items[2].Text := Inttostr(Spritenumber) + ' / ' + Inttostr(Nbp) + ' pieces';
  Result := 'P' + Inttostr(Spritenumber);
End;

Function TFormMain.Testfinal: Boolean;
Var
  A: Integer;
  S: String;

  Function CountPieces: Integer;
  Var
    I: Integer;
    A: Integer;
  Begin
    Log('Count pieces');
    A := 0;
    For I := 1 To Pool.Maxspr Do
      Begin
        If Pool.Gettag(I) = 1 Then
          Inc(A);
      End;
    GetSpriteNumber(A);
    Result := A;
  End;

Begin
  LabelSolved.Visible := False;
  A := CountPieces;
  If A = Pool.Maxspr - 1 Then
    Begin
      Result := True;
      PreviewViewing(False);
      TimerTime.Enabled := False;
      StatusBar.Panels.Items[0].Text := '';
      MenuShowImage.Enabled := True;
      ProgressBar.Position := 0;
      Menugris(True);
      If CountAutoMove > 0 Then
        Begin
          S := 'Congratulations, Jigsaw well done in' + CalcTime;
          If CountAutoMove > Nbp Div 9 Then
            S := S + '  But you cheat.'
          Else
            S := S + '  But you cheat a few.';
        End;
      If Not Completed Then
        Begin
          LabelSolved.Caption := S;
          LabelSolved.Visible := True;
          Completed := True;
        End;
    End
  Else
    Result := False;
End;

Procedure TFormMain.Autodeplace; // The current sprite number is PieceNr
Var
  D: Integer;
  N: Integer;
  Fini: Boolean;
Begin
  Log('AutoDeplace');
  If MoveAuto Then
    Exit;
  Inc(CountMouseMove);
  MoveAuto := True;
  With Pool Do
    Begin
      PieceToRowCol;
      If PartsArray[PieceRow, PieceCol].PRotation > 0 Then
        Begin
          PartsArray[PieceRow, PieceCol].PRotation := 0;
          Setanimstyle(PieceNr, 0);
        End;
      Setpriority(PieceNr, Maxspr + 2);
      If Speed < 3 Then // stepsize (= speed)
        Begin
          Setv(PieceNr, Rnd(-0.3, 0.3), Rnd(-0.5, 0.5));
          D := Distance(Getposx(PieceNr), Getposy(PieceNr), Centrex + Getoffsetx(PieceNr),
            Centrey + Getoffsety(PieceNr)) + 4;
          Case Speed Of
            0:
              D := D Div 4;
            1:
              D := D Div 12;
            2:
              D := D Div 24;
          End;
          If D < 8 Then
            D := 8;
          Gotoxy(PieceNr, Centrex + Getoffsetx(PieceNr), Centrey + Getoffsety(PieceNr), D);
          Setloopctr(PieceNr, D);
          Fini := False;
          Repeat
            For N := 1 To 4 Do // four steps
              Begin
                If assigned(Pool) Then
                  Begin
                    Animatemono(PieceNr);
                    If Getloopctr(PieceNr) = 1 Then
                      Fini := True;
                  End;
                Application.Processmessages;
              End;
          Until Fini;
        End
      Else
        Sleep(40);
      Stop(PieceNr);
      // stops speeds and accelerations
      Setpos(PieceNr, Centrex + Getoffsetx(PieceNr), Centrey + Getoffsety(PieceNr));
      Animatemono(PieceNr);
      Setpriority(PieceNr, 0);
      Settag(PieceNr, 1);
    End;
  Inc(CountAutoMove);
  MoveAuto := False;
End;

Procedure TFormMain.MenuSolutionClick(Sender: TObject);
Var
  Ctr: Integer;
  Nbpbord: Integer;
  // number of border pieces
  Ctrbord: Integer; // count of border pieces right placed
  Minpiece: Integer;
  Maxpiece: Integer;

  Function Pmin(P1, P2: Integer): Integer;
  // give lowest integer
  Begin
    If P1 < P2 Then
      Result := P1
    Else
      Result := P2;
  End;

  Function Pmax(P1, P2: Integer): Integer;
  // give highest integer
  Begin
    If P1 > P2 Then
      Result := P1
    Else
      Result := P2;
  End;

  Function IsborderPiece(P: Integer): Boolean;
  // true when a piece 1..nbp is a border piece
  Begin
    Result := False;
    If (P < Nx) Or (P > Nbp - Nx) Or (P Mod Nx < 2) Then
      Result := True;
  End;

  Procedure SetMenuAutoSolveText;
  Begin
    if IniOptions.AutoSolve = True then
      MenuSolution.Caption := '&Stop'
    else
      MenuSolution.Caption := '&Solution';
  End;

Begin
  Log('Solution');
  IniOptions.AutoSolve := Not IniOptions.AutoSolve;
  SetMenuAutoSolveText;
  // number of border pieces
  If (Nx < 3) Or (Ny < 3) Then
    Nbpbord := Nbp
  Else
    Nbpbord := Nbp - (Nx - 2) * (Ny - 2);
  Ctrbord := 0;
  Minpiece := 1;
  Maxpiece := Nbp;
  For Ctr := 1 To Nbp Do
    If (Pool.Gettag(Ctr) = 1) And IsborderPiece(Ctr) Then
      Inc(Ctrbord);
  With Pool Do
    Begin
      While Testfinal = False Do
        Begin
          Application.Processmessages;
          If IniOptions.AutoSolve = False Then
            Break;
          // random piece choice
          PieceNr := Rndi(Minpiece, Maxpiece);
          Ctr := 0;
          IF IniOptions.StartBorderPieces = False Then
            Ctrbord := Nbpbord + 21;
          IF Ctrbord < Nbpbord Then
            Begin
              While (Gettag(PieceNr) = 1) Or (Ctr > Nbp) Or (IsborderPiece(PieceNr) = False) Do
                Begin
                  Inc(Ctr);
                  Inc(PieceNr);
                  If PieceNr > Maxspr Then
                    PieceNr := 1;
                End;
              Inc(Ctrbord);
            End
          Else
            Begin
              While (Gettag(PieceNr) = 1) Or (Ctr > Nbp) Do
                Begin
                  Inc(Ctr);
                  Inc(PieceNr);
                  If PieceNr > Maxspr Then
                    PieceNr := 1;
                End;
            End;
          Minpiece := Pmin(PieceNr, Minpiece);
          Maxpiece := Pmax(PieceNr, Maxpiece);
          If Not MoveAuto Then
            Autodeplace;
        End;
    End;
  IniOptions.AutoSolve := False;
  SetMenuAutoSolveText;
End;

Procedure TFormMain.MenuHelpAboutClick(Sender: TObject);
Begin
  AboutBox.Showmodal;
End;

End.
