unit uPreview;

interface

uses
{$IF CompilerVersion > 28} // Delphi XE7
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls,
{$ELSE}
  Windows, Messages, SysUtils, Classes, Forms, Graphics, Dialogs,
  // ExtDlgs,
  ExtCtrls, ComCtrls, CommCtrl, Controls, StdCtrls,
{$IFEND}
  Jpeg;

type
  TFormPreview = class(TForm)
    ImagePreview: TImage;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
  public
    procedure CalcImageScaling;
    procedure LoadFromPicture;
    procedure UpdatePosition;
    { Public declarations }
  end;

var
  FormPreview: TFormPreview;
  ImageWidth, ImageHeight: Integer;

implementation

{$R *.dfm}

uses uIniOptions, uMain;

procedure TFormPreview.FormCreate(Sender: TObject);

begin
  BoundsRect := IniOptions.PreviewPlace; // get boundsrect

end;

procedure TFormPreview.FormShow(Sender: TObject);
begin
  UpdatePosition;
end;

procedure TFormPreview.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  IniOptions.PreviewPlace := BoundsRect; // store boundsrect
end;

procedure TFormPreview.CalcImageScaling;
var
  TempImageWidth: Integer;
  TempImageHeight: Integer;

  procedure CalcScaling(var TempW: Integer; var TempH: Integer);
  const
    K = 192;
  begin
    // Bmp0 := Tbitmap.Create;
    // Bmp0.Width := ImagePreview.Picture.Graphic.Width;
    // Bmp0.Height := ImagePreview.Picture.Graphic.Height;
    // Bmp0.Canvas.Draw(0, 0, ImagePreview.Picture.Graphic);

    // adjust image 1 size in the 192 pixels square

    if (TempW > K) or (TempH > K) then
      // scaling
      begin
        if TempW > TempH then
          begin
            // landscape
            TempH := (K * TempH) div TempW;
            TempW := K;
          end
        else
          begin
            // portrait
            TempW := (K * TempW) div TempH;
            TempH := K;
          end;
      end;
    TempH := Trunc(TempH);
    TempW := Trunc(TempW);
  end;

begin
  TempImageWidth := ImageWidth;
  TempImageHeight := ImageHeight;
  CalcScaling(TempImageWidth, TempImageHeight);
  ImagePreview.Width := TempImageWidth;
  ImagePreview.Height := TempImageHeight;
  ImagePreview.Stretch := True;
end;

procedure TFormPreview.LoadFromPicture;
Begin
  ImagePreview.Picture.LoadFromFile(IniOptions.Picture);
  ImageWidth := ImagePreview.Picture.Graphic.Width;
  ImageHeight := ImagePreview.Picture.Graphic.Height;
End;

procedure TFormPreview.UpdatePosition;
begin
  if left < (Formmain.left + Formmain.Width) then
    Begin
      left := Formmain.left + Formmain.Width + 2; // beside mainform
      top := Formmain.top;
    End;
end;

end.
